"""Collect information about the environment and display it.

- Python version
- Package availablility and versions: Zeus, PyTorch, JAX.
- NVIDIA GPU availability: Number of GPUs and moels.
- AMD GPU availability: Number of GPUs and models.
- Intel RAPL availability: Number of CPUs and whether DRAM measurements are available.
"""

from __future__ import annotations

import platform
import shutil

import zeus
from zeus.device.cpu.rapl import RAPLCPU, ZeusdRAPLCPU
from zeus.device.exception import ZeusBaseCPUError, ZeusBaseGPUError, ZeusBaseSoCError
from zeus.utils import framework
from zeus.device import get_gpus, get_cpus, get_soc
from zeus.device.cpu import RAPLCPUs
from zeus.device.gpu.common import ZeusGPUInitError, EmptyGPUs
from zeus.device.cpu.common import ZeusCPUInitError, EmptyCPUs
from zeus.device.soc.common import ZeusSoCInitError, EmptySoC
from zeus.device.soc.apple import AppleSilicon
from zeus.device.soc.jetson import Jetson


SECTION_SEPARATOR = "-" * shutil.get_terminal_size().columns


def show_env():
    """Collect information about the environment and display it."""
    print(SECTION_SEPARATOR)
    print("## Package availability and versions\n")
    print("Logging output:")

    package_availability = "  Python version: " + platform.python_version() + "\n"
    package_availability += f"  Zeus: {zeus.__version__}\n"

    try:
        torch_available = framework.torch_is_available()
        torch_cuda_available = True
    except RuntimeError:
        torch_available = framework.torch_is_available(ensure_cuda=False)
        torch_cuda_available = False

    if torch_available and torch_cuda_available:
        torch = framework.MODULE_CACHE["torch"]
        package_availability += f"  PyTorch: {torch.__version__} (with CUDA support)\n"
    elif torch_available and not torch_cuda_available:
        torch = framework.MODULE_CACHE["torch"]
        package_availability += (
            f"  PyTorch: {torch.__version__} (without CUDA support)\n"
        )
    else:
        package_availability += "  PyTorch: not available\n"

    try:
        jax_available = framework.jax_is_available()
        jax_cuda_available = True
    except RuntimeError:
        jax_available = framework.jax_is_available(ensure_cuda=False)
        jax_cuda_available = False

    if jax_available and jax_cuda_available:
        jax = framework.MODULE_CACHE["jax"]
        package_availability += f"  JAX: {jax.__version__} (with CUDA support)\n"
    elif jax_available and not jax_cuda_available:
        jax = framework.MODULE_CACHE["jax"]
        package_availability += f"  JAX: {jax.__version__} (without CUDA support)\n"
    else:
        package_availability += "  JAX: not available\n"

    print("\nDetected:\n" + package_availability)

    print(SECTION_SEPARATOR)
    print("## Device availability\n")
    print("Logging output:")

    gpu_availability = ""
    try:
        gpus = get_gpus()
    except ZeusGPUInitError:
        gpus = EmptyGPUs()
    except ZeusBaseGPUError as e:
        gpu_availability += f"  Error initializing GPUs: {e}\n"
        gpus = EmptyGPUs()
    if len(gpus) > 0:
        for i in range(len(gpus)):
            gpu_availability += f"  GPU {i}: {gpus.getName(i)}\n"
    else:
        gpu_availability += "  No GPUs available.\n"
    print("\nDetected:\n" + gpu_availability)

    print(SECTION_SEPARATOR)
    print("## CPU availability\n")
    print("Logging output:")

    cpu_availability = ""
    try:
        cpus = get_cpus()
    except ZeusCPUInitError:
        cpus = EmptyCPUs()
    except ZeusBaseCPUError as e:
        cpu_availability += f"  Error initializing CPUs: {e}\n"
        cpus = EmptyCPUs()

    if len(cpus) > 0:
        assert isinstance(cpus, RAPLCPUs)
        for i, cpu in enumerate(cpus.cpus):
            if isinstance(cpu, ZeusdRAPLCPU):
                cpu_availability += f"  CPU {i}:\n    CPU measurements available (Zeusd at {cpu.zeusd_sock_path})\n"
                if cpu.supportsGetDramEnergyConsumption():
                    cpu_availability += f"    DRAM measurements available (Zeusd at {cpu.zeusd_sock_path})\n"
                else:
                    cpu_availability += "    DRAM measurements unavailable\n"
            elif isinstance(cpu, RAPLCPU):
                cpu_availability += f"  CPU {i}:\n    CPU measurements available ({cpu.rapl_file.path})\n"
                if cpu.supportsGetDramEnergyConsumption():
                    dram = cpu.dram
                    assert dram is not None
                    cpu_availability += (
                        f"    DRAM measurements available ({dram.path})\n"
                    )
                else:
                    cpu_availability += "    DRAM measurements unavailable\n"
            else:
                raise TypeError("Unexpected CPU type: " + str(type(cpu)))
    else:
        cpu_availability += "  No CPUs available.\n"
    print("\nDetected:\n" + cpu_availability)

    print(SECTION_SEPARATOR)
    print("## SoC availability\n")
    print("Logging output:")

    soc_availability = ""
    try:
        soc = get_soc()
    except ZeusSoCInitError:
        soc = EmptySoC()
    except ZeusBaseSoCError as e:
        soc_availability += f"  Error initializing SoC: {e}\n"
        soc = EmptySoC()

    if isinstance(soc, AppleSilicon):
        metrics = soc.getAvailableMetrics()
        soc_availability += "  Apple Silicon SoC available.\n"
        soc_availability += f"  Available metrics: {', '.join(metrics)}\n"
    elif isinstance(soc, Jetson):
        metrics = soc.getAvailableMetrics()
        soc_availability += "  NVIDIA Jetson SoC available.\n"
        soc_availability += f"  Available metrics: {', '.join(metrics)}\n"
    else:
        soc_availability += "  No SoC available.\n"

        # If this is an Apple Silicon device but AppleSilicon was not detected,
        # print out a warning.
        if platform.system() == "Darwin" and platform.machine() in ["arm64", "aarch64"]:
            soc_availability += (
                "\nThis appears to be an Apple Silicon device, but it wasn't picked up by Zeus.\n"
                "Have you installed Zeus with the `apple` extra?\n\n"
                "    pip install 'zeus[apple]'\n"
            )
    print("\nDetected:\n" + soc_availability)

    print(SECTION_SEPARATOR)


if __name__ == "__main__":
    show_env()



"""Electricity price providers used for price-aware optimizers."""

from __future__ import annotations

import abc
from dataclasses import dataclass
from enum import Enum
import queue
import requests
import json
import multiprocessing as mp

from typing import Literal
from datetime import datetime, timezone, timedelta
from collections import defaultdict

from zeus.exception import ZeusBaseError
from zeus.monitor import ZeusMonitor
from zeus.utils.logging import get_logger

logger = get_logger(__name__)


def get_time_info() -> tuple[str, str, int]:
    """Retrieve the month, day_type (weekend or weekday), and hour."""
    now = datetime.now()

    month = now.strftime("%B")

    day_type = "Weekend" if now.weekday() >= 5 else "Weekday"

    hour = now.hour

    return month, day_type, hour


class ZeusElectricityPriceHTTPError(ZeusBaseError):
    """Exception when HTTP request to electricity price provider fails."""

    def __init__(self, message: str) -> None:
        """Initialize HTTP request exception."""
        super().__init__(message)


class ZeusElectricityPriceNotFoundError(ZeusBaseError):
    """Exception when electricity price measurement could not be retrieved."""

    def __init__(self, message: str) -> None:
        """Initialize price not found exception."""
        super().__init__(message)


class ElectricityPriceProvider(abc.ABC):
    """Abstract class for implementing ways to fetch electricity price."""

    @abc.abstractmethod
    def get_current_electricity_prices(self) -> dict[str, list]:
        """Abstract method for fetching the current electricity price of the set location of the class."""
        pass


class OpenEIClient(ElectricityPriceProvider):
    """Electricity Price Provider with OpenEI API.

    Reference:

    1. [OpenEI](https://openei.org/wiki/Main_Page)
    2. [OpenEI Utility Rates API](https://apps.openei.org/services/doc/rest/util_rates/?version=7)
    """

    def __init__(
        self,
        location: tuple[float, float],
        label: str,
        sector: Literal[
            "Residential", "Commercial", "Industrial", "Lighting"
        ] = "Residential",
        radius: int = 0,
    ) -> None:
        """Initializes OpenEI Utility Rates Provider.

        Args:
            location: tuple of latitude and longitude (latitude, longitude)
            label: unique identifier of a particular variant of a utility company's rate
            sector: depends on which sector of electricity is relevant to you
            radius: search radius for utility rates from the location
        """
        self.lat, self.long = location
        self.label = label
        self.sector = sector
        self.radius = radius

    def search_json(self, data, key_name, target_value, return_value):
        """Recursively search for a key in a nested JSON and return the return_value field if found."""
        results = []

        if isinstance(data, dict):
            for key, val in data.items():
                # Check if the current dictionary contains the matching key-value pair
                if key == key_name and val == target_value:
                    # If "energyratestructure" exists at the same level, add it to results
                    if return_value in data:
                        results.append(data[return_value])
                    else:
                        results.append(None)

                # Recursively search deeper in nested dictionaries
                results.extend(
                    self.search_json(val, key_name, target_value, return_value)
                )

        elif isinstance(data, list):
            for item in data:
                results.extend(
                    self.search_json(item, key_name, target_value, return_value)
                )

        return results

    def get_current_electricity_prices(self) -> dict[str, list]:
        """Fetches current carbon intensity of the location of the class."""
        try:
            url = (
                "https://api.openei.org/utility_rates?version=latest&format=json"
                + f"&api_key=tJASWWgPhBRpiZCwfhtKV2A3gyNxbDfvQvdI5Wa7&lat={self.lat}"
                + f"&lon={self.long}&radius={self.radius}"
                + f"&detail=full&sector={self.sector}"
            )
            resp = requests.get(url)
            data = resp.json()

        except requests.exceptions.RequestException as e:
            raise ZeusElectricityPriceHTTPError(
                f"Failed to retrieve current electricity price measurement: {e}"
            ) from e

        try:
            if "label" not in json.dumps(data):
                raise ZeusElectricityPriceNotFoundError(
                    f"No rates found for lat, lon: [{self.lat}, {self.long}]."
                )

            energy_rate_structure = self.search_json(
                data, "label", self.label, "energyratestructure"
            )
            energy_weekday_schedule = self.search_json(
                data, "label", self.label, "energyweekdayschedule"
            )
            energy_weekend_schedule = self.search_json(
                data, "label", self.label, "energyweekendschedule"
            )

            if (
                not energy_rate_structure
                or not energy_weekday_schedule
                or not energy_weekend_schedule
            ):
                raise ZeusElectricityPriceNotFoundError(
                    f"No rates found for the label: {self.label}."
                )

            rate_data = {
                "energy_rate_structure": energy_rate_structure[0],
                "energy_weekday_schedule": energy_weekday_schedule[0],
                "energy_weekend_schedule": energy_weekend_schedule[0],
            }
            return rate_data

        except (KeyError, ValueError) as e:
            logger.error(
                "Error occurred while processing electricity price data: %s", e
            )
            raise ZeusElectricityPriceNotFoundError(
                "Failed to process electricity price data."
            ) from e


@dataclass
class EnergyCostMeasurement:
    """Measurement result of one window.

    Attributes:
        time: Time elapsed (in seconds) during the measurement window.
        gpu_energy: Maps GPU indices to the energy consumed (in Joules) during the
            measurement window. GPU indices are from the DL framework's perspective
            after applying `CUDA_VISIBLE_DEVICES`.
        gpu_energy_cost: Maps GPU indices to the electricity cost (in $) during the
            measurement window. GPU indices are from the DL framework's perspective
            after applying `CUDA_VISIBLE_DEVICES`.
        cpu_energy: Maps CPU indices to the energy consumed (in Joules) during the measurement
            window. Each CPU index refers to one powerzone exposed by RAPL (intel-rapl:d). This can
            be 'None' if CPU measurement is not available.
        cpu_energy_cost: Maps CPU indices to the electricity cost (in $) during the measurement
            window. Each CPU index refers to one powerzone exposed by RAPL (intel-rapl:d). This can
            be 'None' if CPU measurement is not available.
        dram_energy: Maps CPU indices to the energy consumed (in Joules) during the measurement
            window. Each CPU index refers to one powerzone exposed by RAPL (intel-rapl:d) and DRAM
            measurements are taken from sub-packages within each powerzone. This can be 'None' if
            CPU measurement is not available or DRAM measurement is not available.
        dram_energy_cost: Maps CPU indices to the electricity cost (in $) during the measurement
            window. Each CPU index refers to one powerzone exposed by RAPL (intel-rapl:d). This can be 'None' if
            CPU measurement is not available or DRAM measurement is not available.
    """

    time: float
    gpu_energy: dict[int, float]
    gpu_energy_cost: dict[int, float]
    cpu_energy: dict[int, float] | None = None
    cpu_energy_cost: dict[int, float] | None = None
    dram_energy: dict[int, float] | None = None
    dram_energy_cost: dict[int, float] | None = None


class Op(Enum):
    """Enum used to communicate between EnergyCostMonitor and _polling_process."""

    BEGIN = 0
    END = 1
    NEXTITER = 2


class EnergyCostMonitor:
    """Measure the energy, energy cost, and time consumption of a block of code.

    Works for multi-GPU and heterogeneous GPU types. Aware of `CUDA_VISIBLE_DEVICES`.
    For instance, if `CUDA_VISIBLE_DEVICES=2,3`, GPU index `1` passed into `gpu_indices`
    will be interpreted as CUDA device `3`.

    You can mark the beginning and end of a measurement window, during which the energy cost,
    GPU energy, and time consumed will be recorded. Multiple concurrent measurement windows
    are supported.
    """

    def __init__(
        self,
        electricity_price_provider: ElectricityPriceProvider,
        gpu_indices: list[int] | None = None,
        cpu_indices: list[int] | None = None,
        sync_execution_with: Literal["torch", "jax"] = "torch",
    ) -> None:
        """Initializes Energy Cost Monitor.

        Args:
            electricity_price_provider: provider for which electricity price values will be fetched from
            gpu_indices: Indices of all the CUDA devices to monitor. Time/Energy measurements
                will begin and end at the same time for these GPUs (i.e., synchronized).
                If None, all the GPUs available will be used. `CUDA_VISIBLE_DEVICES`
                is respected if set, e.g., GPU index `1` passed into `gpu_indices` when
                `CUDA_VISIBLE_DEVICES=2,3` will be interpreted as CUDA device `3`.
                `CUDA_VISIBLE_DEVICES`s formatted with comma-separated indices are supported.
            cpu_indices: Indices of the CPU packages to monitor. If None, all CPU packages will
                be used.
            sync_execution_with: Deep learning framework to use to synchronize CPU/GPU computations.
                Defaults to `"torch"`, in which case `torch.cuda.synchronize` will be used.
                See [`sync_execution`][zeus.utils.framework.sync_execution] for more details.
        """
        self.zeus_monitor = ZeusMonitor(
            gpu_indices=gpu_indices,
            cpu_indices=cpu_indices,
            sync_execution_with=sync_execution_with,
            approx_instant_energy=True,
        )
        self.electricity_price_provider = electricity_price_provider
        self.current_keys = set()

        # set up process and shared queues
        self.context = mp.get_context("spawn")
        self.command_q = self.context.Queue()
        self.finished_q = self.context.Queue()

    def begin_window(self, key: str, sync_execution: bool = True) -> None:
        """Begin a new measurement window.

        Args:
            key: Unique name of the measurement window.
            sync_execution: Whether to wait for asynchronously dispatched computations
                to finish before starting the measurement window. For instance, PyTorch
                and JAX will run GPU computations asynchronously, and waiting them to
                finish is necessary to ensure that the measurement window captures all
                and only the computations dispatched within the window.
        """
        # check if key is already used
        if key in self.current_keys:
            raise ValueError(f"Measurement window '{key}' already exists")
        self.current_keys.add(key)

        # start window
        self.zeus_monitor.begin_window(key, sync_execution=sync_execution)

        # if there were previously no active windows, start polling process
        if len(self.current_keys) == 1:
            self.polling_process = self.context.Process(
                target=_polling_process,
                args=(
                    self.command_q,
                    self.finished_q,
                    self.zeus_monitor.gpu_indices,
                    self.zeus_monitor.cpu_indices,
                    self.electricity_price_provider,
                ),
            )
            self.polling_process.start()

        # start subwindows
        self.command_q.put((Op.BEGIN, key))

    def end_window(
        self, key: str, sync_execution: bool = True
    ) -> EnergyCostMeasurement:
        """End a measurement window and return the time, energy consumption, and energy cost.

        Args:
            key: Name of an active measurement window.
            sync_execution: Whether to wait for asynchronously dispatched computations
                to finish before starting the measurement window. For instance, PyTorch
                and JAX will run GPU computations asynchronously, and waiting them to
                finish is necessary to ensure that the measurement window captures all
                and only the computations dispatched within the window.
        """
        # check if begin_window has been called with key before
        if key not in self.current_keys:
            raise ValueError(f"Measurement window '{key}' does not exist")

        # end window
        self.command_q.put((Op.END, key))
        (
            gpu_energy_cost,
            cpu_energy_cost,
            dram_energy_cost,
        ) = self.finished_q.get()
        self.current_keys.remove(key)

        overall_measurement = self.zeus_monitor.end_window(
            key, sync_execution=sync_execution
        )

        measurement = EnergyCostMeasurement(
            time=overall_measurement.time,
            gpu_energy=overall_measurement.gpu_energy,
            cpu_energy=overall_measurement.cpu_energy,
            dram_energy=overall_measurement.dram_energy,
            gpu_energy_cost=gpu_energy_cost,
            cpu_energy_cost=cpu_energy_cost or None,
            dram_energy_cost=dram_energy_cost or None,
        )

        return measurement


def _polling_process(
    command_q: mp.Queue,
    finished_q: mp.Queue,
    gpu_indices: list[int],
    cpu_indices: list[int],
    electricity_price_provider: ElectricityPriceProvider,
):
    index = 0
    zeus_monitor = ZeusMonitor(
        gpu_indices=gpu_indices,
        cpu_indices=cpu_indices,
    )
    gpu_energy_cost = defaultdict(lambda: defaultdict(float))
    cpu_energy_cost = defaultdict(lambda: defaultdict(float))
    dram_energy_cost = defaultdict(lambda: defaultdict(float))
    energy_measurements = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
    keys = set()

    # Fetch electricity price data
    try:
        electricity_price_data = (
            electricity_price_provider.get_current_electricity_prices()
        )
        energy_rate_structure = electricity_price_data["energy_rate_structure"]
        energy_weekday_schedule = electricity_price_data["energy_weekday_schedule"]
        energy_weekend_schedule = electricity_price_data["energy_weekend_schedule"]
    except Exception as e:
        logger.error("Failed to retrieve electricity price: %s.", e)
        return

    # record energy measurement
    def _update_energy_measurements(key: str, datetime: datetime):
        measurement = zeus_monitor.end_window(key, sync_execution=False)

        for gpu_index, energy_measurement in measurement.gpu_energy.items():
            energy_measurements[key][datetime][f"gpu_{gpu_index}"] = energy_measurement

        if measurement.cpu_energy:
            for cpu_index, energy_measurement in measurement.cpu_energy.items():
                energy_measurements[key][datetime][
                    f"cpu_{cpu_index}"
                ] = energy_measurement

        if measurement.dram_energy:
            for dram_index, energy_measurement in measurement.dram_energy.items():
                energy_measurements[key][datetime][
                    f"dram_{dram_index}"
                ] = energy_measurement

    # update cumulative electricity costs
    def _update_energy_costs(key: str):
        for dt, measurement_map in energy_measurements[key].items():
            for index, energy in measurement_map.items():
                hardware_type, num_index = index.split("_")

                month = dt.month - 1
                hour = dt.hour
                day_of_week = dt.weekday()

                tier = (
                    energy_weekday_schedule[month][hour]
                    if day_of_week < 5
                    else energy_weekend_schedule[month][hour]
                )

                try:
                    flat_rate = energy_rate_structure[tier][0]["rate"]
                except (IndexError, KeyError, TypeError):
                    logger.error("Failed to parse electricity rate structure.")
                    return

                cost = (
                    energy / 3.6e6
                ) * flat_rate  # Convert Wh to kWh and multiply by rate

                if hardware_type == "gpu":
                    gpu_energy_cost[key][int(num_index)] += cost
                elif hardware_type == "cpu":
                    cpu_energy_cost[key][int(num_index)] += cost
                elif hardware_type == "dram":
                    dram_energy_cost[key][int(num_index)] += cost

        del energy_measurements[key]

    while True:
        # calculate current time
        now = datetime.now(timezone.utc)
        hour_floor = now.replace(minute=0, second=0, microsecond=0)
        hour_ceil = hour_floor + timedelta(hours=1)

        # start windows
        for key in keys:
            zeus_monitor.begin_window(key, sync_execution=False)

        try:
            # continuously fetch from command q until hit hour
            while now < hour_ceil:
                seconds_until_hour = (hour_ceil - now).total_seconds()
                op, key = command_q.get(timeout=seconds_until_hour)

                if op == Op.BEGIN:
                    zeus_monitor.begin_window(key, sync_execution=False)
                    keys.add(key)
                elif op == Op.END:
                    # update if has not been updated in a while
                    _update_energy_measurements(key, hour_floor)
                    _update_energy_costs(key)
                    finished_q.put(
                        (
                            dict(gpu_energy_cost.pop(key)),
                            dict(cpu_energy_cost.pop(key)),
                            dict(dram_energy_cost.pop(key)),
                        )
                    )
                    keys.remove(key)

                    if len(keys) == 0:
                        return
                elif op == Op.NEXTITER:
                    # for testing purposes only, force monitor to move onto next hour
                    # this op will never be sent into command q outside of testing environments
                    break

                now = datetime.now(timezone.utc)
        except queue.Empty:
            # if nothing from finished_q.get(), continue
            pass

        # record energy values
        index += 1
        for key in keys:
            _update_energy_measurements(key, hour_floor)



"""Zeus batch size optimizer service layer."""

from __future__ import annotations

import json
from datetime import datetime
from typing import Any

import numpy as np
from numpy.random import Generator as np_Generator
from sqlalchemy.ext.asyncio.session import AsyncSession
from zeus.optimizer.batch_size.server.batch_size_state.commands import (
    CreateConcurrentTrial,
    CreateExplorationTrial,
    CreateMabTrial,
    CreateTrial,
    ReadTrial,
    UpdateTrial,
)
from zeus.optimizer.batch_size.server.batch_size_state.models import (
    BatchSizeBase,
    ExplorationsPerJob,
    GaussianTsArmState,
    Trial,
    TrialResultsPerBs,
)
from zeus.optimizer.batch_size.server.batch_size_state.repository import (
    BatchSizeStateRepository,
)
from zeus.optimizer.batch_size.server.database.schema import TrialStatus, TrialType
from zeus.optimizer.batch_size.server.exceptions import (
    ZeusBSOServiceBadOperationError,
    ZeusBSOValueError,
)
from zeus.optimizer.batch_size.server.job.commands import (
    CreateJob,
    UpdateExpDefaultBs,
    UpdateGeneratorState,
    UpdateJobMinCost,
    UpdateJobStage,
)
from zeus.optimizer.batch_size.server.job.models import JobState
from zeus.optimizer.batch_size.server.job.repository import JobStateRepository
from zeus.optimizer.batch_size.server.services.commands import (
    GetNormal,
    GetRandomChoices,
    UpdateArm,
)
from zeus.utils.metric import zeus_cost


class ZeusService:
    """Zeus Service that interacts with database using repository.

    Provides application layer methods to communicate with database.
    Each method is one or more number of db operations that have to be done at the same time.
    """

    def __init__(self, db_session: AsyncSession):
        """Set up repositories to use to talk to database."""
        self.bs_repo = BatchSizeStateRepository(db_session)
        self.job_repo = JobStateRepository(db_session)

    async def get_arms(self, job_id: str) -> list[GaussianTsArmState]:
        """Get GaussianTs arm states for all arms(job_id, batch size).

        Args:
            job_id: Job id

        Returns:
            list of arms
        """
        return await self.bs_repo.get_arms(job_id)

    async def get_arm(self, bs: BatchSizeBase) -> GaussianTsArmState | None:
        """Get arm state for one arm.

        Args:
            bs: (job_id, batch size) pair that represents one arm

        Returns:
            Result arm state or None if we cannot find that arm
        """
        return await self.bs_repo.get_arm(bs)

    async def get_explorations_of_job(self, job_id: str) -> ExplorationsPerJob:
        """Get all explorations we have done for that job.

        Args:
            job_id: Job id

        Returns:
            list of explorations per each batch size
        """
        return await self.bs_repo.get_explorations_of_job(job_id)

    def update_trial(self, updated_trial: UpdateTrial) -> None:
        """Update trial.

        (1) update the corresponding trial.
        (2) we update the min training cost observed so far if we have to.

        Args:
            updated_trial: Result of training that batch size

        Raises:
            [`ZeusBSOServiceBadOperationError`][zeus.optimizer.batch_size.server.exceptions.ZeusBSOServiceBadOperationError]: When we didn't fetch the job or trial during this session. This operation should have
                    fetched the job and trial first. Also, check if trial type is matching with fetched trial's type.
        """
        trial = self._get_trial(
            ReadTrial(
                job_id=updated_trial.job_id,
                batch_size=updated_trial.batch_size,
                trial_number=updated_trial.trial_number,
            )
        )
        if trial.status != TrialStatus.Dispatched:
            raise ZeusBSOServiceBadOperationError("Trial already has a result.")

        self.bs_repo.updated_current_trial(updated_trial)

        # Update the corresponding batch size's min cost if needed.
        if updated_trial.status != TrialStatus.Failed:
            job = self._get_job(updated_trial.job_id)
            if updated_trial.energy is None or updated_trial.time is None:
                raise ZeusBSOValueError(
                    "Energy and time should be set if the trial is not failed."
                )
            cur_cost = zeus_cost(
                updated_trial.energy, updated_trial.time, job.eta_knob, job.max_power
            )
            if job.min_cost is None or job.min_cost > cur_cost:
                self.job_repo.update_min(
                    UpdateJobMinCost(
                        job_id=job.job_id,
                        min_cost=cur_cost,
                        min_cost_batch_size=updated_trial.batch_size,
                    )
                )

    def update_arm_state(
        self,
        arm: UpdateArm,
    ) -> None:
        """Update arm state.

        Args:
            arm: Updated arm state.

        Raises:
            `ZeusBSOServiceBadOperationError`: When we didn't fetch the job or trial during this session. This operation should have
                    fetched the job and trial first. Also, check if trial type is matching with fetched trial's type.
        """
        self._check_job_fetched(arm.trial.job_id)
        trial = self._get_trial(
            ReadTrial(
                job_id=arm.trial.job_id,
                batch_size=arm.trial.batch_size,
                trial_number=arm.trial.trial_number,
            )
        )
        if trial.type != TrialType.MAB:
            raise ZeusBSOServiceBadOperationError(
                "Cannot update an arm since this trial is not issued from MAB stage."
            )
        self.bs_repo.update_arm_state(arm.updated_arm)

    def update_exp_default_bs(self, updated_default_bs: UpdateExpDefaultBs) -> None:
        """Update the default batch size for exploration.

        Args:
            updated_default_bs: Job Id and new default batch size

        Raises:
            `ZeusBSOServiceBadOperationError`: When we didn't fetch the job during this session. This operation should have
                    fetched the job first.
        """
        self._check_job_fetched(updated_default_bs.job_id)
        self.job_repo.update_exp_default_bs(updated_default_bs)

    async def create_trial(
        self, trial: CreateExplorationTrial | CreateMabTrial | CreateConcurrentTrial
    ) -> ReadTrial:
        """Create a new trial.

        Args:
            trial: New trial to create.

        Raises:
            `ZeusBSOServiceBadOperationError`: When we didn't fetch the job during this session. This operation should have
                    fetched the job first.
        """
        self._check_job_fetched(trial.job_id)
        trial_number = await self.bs_repo.get_next_trial_number(trial.job_id)
        self.bs_repo.create_trial(
            CreateTrial(**trial.dict(), trial_number=trial_number)
        )
        return ReadTrial(
            job_id=trial.job_id, batch_size=trial.batch_size, trial_number=trial_number
        )

    def get_random_choices(self, choice: GetRandomChoices) -> np.ndarray[Any, Any]:
        """Get randome choices based on job's seed.

        If seed is not None (set by the user) we get the random choices from the generator that is stored in the database.
        Otherwise, we get random choices based on random seed.

        Args:
            choice: Job id and list of choices

        Returns:
            reuslt random choices

        Raises:
            `ZeusBSOServiceBadOperationError`: When we didn't fetch the job during this session. This operation should have
                    fetched the job first.
        """
        arr = np.array(choice.choices)
        rng, should_update = self._get_generator(choice.job_id)
        res = rng.choice(arr, len(arr), replace=False)

        if should_update:
            # If we used the generator from database, should update the generator state after using it
            self.job_repo.update_generator_state(
                UpdateGeneratorState(
                    job_id=choice.job_id, state=json.dumps(rng.__getstate__())
                )
            )

        return res

    def get_normal(self, arg: GetNormal) -> float:
        """Sample from normal distribution and update the generator state if seed was set.

        Args:
            arg: args for `numpy.random.normal`, which is loc(mean of distribution) and scale(stdev of distribution)

        Returns:
            Drawn sample.

        Raises:
            `ZeusBSOServiceBadOperationError`: When we didn't fetch the job during this session. This operation should have
                    fetched the job first.
        """
        rng, should_update = self._get_generator(arg.job_id)
        res = rng.normal(arg.loc, arg.scale)

        if should_update:
            # If we used the generator from database, should update the generator state after using it
            self.job_repo.update_generator_state(
                UpdateGeneratorState(
                    job_id=arg.job_id, state=json.dumps(rng.__getstate__())
                )
            )

        return res

    async def get_job(self, job_id: str) -> JobState | None:
        """Get job from database.

        Args:
            job_id: Job Id

        Returns:
            JobState if we found one, None if we couldn't find a job matching the job id.
        """
        return await self.job_repo.get_job(job_id)

    async def get_trial(self, trial: ReadTrial) -> Trial | None:
        """Get a trial from database.

        Args:
            trial: (Job Id, batch size, trial_number) triplet.

        Returns:
            Trial if we found one, None if we couldn't find a job matching trial.
        """
        return await self.bs_repo.get_trial(trial)

    def create_job(self, new_job: CreateJob) -> None:
        """Create a new job.

        Args:
            new_job: Configuration of a new job
        """
        return self.job_repo.create_job(new_job)

    async def get_trial_results_of_bs(self, bs: BatchSizeBase) -> TrialResultsPerBs:
        """Load window size amount of results for a given batch size. If window size <= 0, load all of them.

        Args:
            bs: (job_id, batch size) pair.

        Returns:
            list of windowed measurements in descending order for that (job_id, batch size)

        Raises:
            `ZeusBSOServiceBadOperationError`: When we didn't fetch the job during this session. This operation should have
                    fetched the job first.
        """
        job = self._get_job(bs.job_id)
        return await self.bs_repo.get_trial_results_of_bs(
            BatchSizeBase(job_id=bs.job_id, batch_size=bs.batch_size),
            job.window_size,
        )

    def create_arms(self, new_arms: list[GaussianTsArmState]) -> None:
        """Create GuassianTs arms for the job.

        Args:
            new_arms: List of new arm states

        Raises:
            `ZeusBSOServiceBadOperationError`: When we didn't fetch the job during this session. This operation should have
                    fetched the job first.
        """
        if len(new_arms) != 0:
            self._check_job_fetched(new_arms[0].job_id)
            self.bs_repo.create_arms(new_arms)

    def update_job_stage(self, updated_stage: UpdateJobStage) -> None:
        """Update the job stage (Pruning -> MAB).

        Args:
            updated_stage: Updated stage.

        Raises:
            `ZeusBSOServiceBadOperationError`: When we didn't fetch the job during this session. This operation should have
                    fetched the job first.
        """
        self._check_job_fetched(updated_stage.job_id)
        self.job_repo.update_stage(updated_stage)

    async def delete_job(self, job_id: str) -> bool:
        """Delete the job.

        Args:
            job_id: ID of the job.

        Returns:
            True if the job is deleted. False if none was deleted
        """
        return await self.job_repo.delete_job(job_id)

    def _get_generator(self, job_id: str) -> tuple[np_Generator, bool]:
        """Get generator based on job_id. If mab_seed is not none, we should update the state after using generator.

        Returns:
            Tuple of [Generator, if we should update state]
        """
        job_state = self._get_job(job_id)

        rng = np.random.default_rng(int(datetime.now().timestamp()))

        should_update = job_state.mab_seed is not None
        if job_state.mab_seed is not None:
            if job_state.mab_random_generator_state is None:
                raise ZeusBSOValueError(
                    "Seed is set but generator state is none. Should be impossible"
                )

            state = json.loads(job_state.mab_random_generator_state)
            rng.__setstate__(state)

        return (rng, should_update)

    def _get_job(self, job_id: str) -> JobState:
        """Get the job from the session. If we couldn't find the job, raise a `ZeusBSOServiceBadOperationError`."""
        res = self.job_repo.get_job_from_session(job_id)
        if res is None:
            raise ZeusBSOServiceBadOperationError(
                f"Should have fetched the job first or job does not exist(job_id = {job_id})"
            )
        return res

    def _get_trial(self, trial: ReadTrial) -> Trial:
        """Get the job from the session. If we couldn't find the trial, raise a `ZeusBSOServiceBadOperationError`."""
        res = self.bs_repo.get_trial_from_session(trial)
        if res is None:
            raise ZeusBSOServiceBadOperationError(
                f"Should have fetched the trial first or trial does not exist(trial = {trial})"
            )
        return res

    def _check_job_fetched(self, job_id: str) -> None:
        """Check if we fetched the job in the current session. If we didn't raise a `ZeusBSOServiceBadOperationError`."""
        if not self.job_repo.check_job_fetched(job_id):
            raise ZeusBSOServiceBadOperationError(
                f"check_job_fetched: {job_id} is not currently in the session"
            )



"""Tools related to environment variables."""

from __future__ import annotations

import os
from typing import Type, TypeVar, cast

T = TypeVar("T")


def get_env(name: str, valtype: Type[T], default: T | None = None) -> T:
    """Fetch an environment variable and cast it to the given type."""
    try:
        if valtype is bool:
            val = os.environ[name].lower()
            if val not in ["true", "false"]:
                raise ValueError(f"Strange boolean environment variable value '{val}'")
            return cast(T, val == "true")
        return valtype(os.environ[name])  # type: ignore
    except KeyError:
        if default is not None:
            return default
        raise ValueError(f"Missing environment variable '{name}'") from None



"""Track and export energy and power metrics via Prometheus."""

from __future__ import annotations

import abc
import time
import warnings
from typing import Sequence
import multiprocessing as mp
from multiprocessing.context import SpawnProcess
from dataclasses import dataclass

from prometheus_client import (
    CollectorRegistry,
    Histogram,
    Counter,
    Gauge,
    push_to_gateway,
)

from zeus.monitor.power import PowerMonitor
from zeus.monitor.energy import ZeusMonitor
from zeus.utils.framework import sync_execution as sync_execution_fn
from zeus.device.cpu import get_cpus


@dataclass
class MonitoringProcessState:
    """Represents the state of a monitoring window."""

    queue: mp.Queue
    proc: SpawnProcess


class Metric(abc.ABC):
    """Abstract base class for all metric types in Zeus.

    Defines a common interface for metrics, ensuring consistent behavior
    for `begin_window` and `end_window` operations.
    """

    @abc.abstractmethod
    def begin_window(self, name: str, sync_execution: bool = True) -> None:
        """Start a new measurement window.

        Args:
            name (str): Name of the measurement window.
            sync_execution (bool): Whether to wait for asynchronously dispatched computations
                to finish before starting the measurement window.
        """
        pass

    @abc.abstractmethod
    def end_window(self, name: str, sync_execution: bool = True) -> None:
        """End the current measurement window and report metrics.

        Args:
            name (str): Name of the measurement window.
            sync_execution (bool): Whether to wait for asynchronously dispatched computations
                to finish before starting the measurement window.
        """
        pass


class EnergyHistogram(Metric):
    """Measures the energy consumption a code range and exports a histogram metrics.

    Tracks energy consumption for GPUs, CPUs, and DRAM as Prometheus Histogram metrics.
    """

    def __init__(
        self,
        cpu_indices: list,
        gpu_indices: list,
        pushgateway_url: str,
        job: str,
        gpu_bucket_range: Sequence[float] = [50.0, 100.0, 200.0, 500.0, 1000.0],
        cpu_bucket_range: Sequence[float] = [10.0, 50.0, 100.0, 500.0, 1000.0],
        dram_bucket_range: Sequence[float] = [5.0, 10.0, 20.0, 50.0, 150.0],
    ) -> None:
        """Initialize the EnergyHistogram class.

        Sets up the Prometheus Histogram metrics to track energy consumption for GPUs, CPUs, and DRAMs.
        The data will be collected and pushed to the Prometheus Push Gateway at regular intervals.

        Args:
            cpu_indices (list): List of CPU indices to monitor.
            gpu_indices (list): List of GPU indices to monitor.
            pushgateway_url (str): URL of the Prometheus Push Gateway where metrics will be pushed.
            job (str): Name of the Prometheus job to associate with the energy metrics.
            gpu_bucket_range (list[float], optional): Bucket ranges for GPU energy histograms.
                Defaults to [50.0, 100.0, 200.0, 500.0, 1000.0].
            cpu_bucket_range (list[float], optional): Bucket ranges for CPU energy histograms.
                Defaults to [10.0, 20.0, 50.0, 100.0, 200.0].
            dram_bucket_range (list[float], optional): Bucket ranges for DRAM energy histograms.
                Defaults to [5.0, 10.0, 20.0, 50.0, 150.0].

        Raises:
            ValueError: If any of the bucket ranges (GPU, CPU, DRAM) is an empty list.
        """
        self.gpu_bucket_range = gpu_bucket_range
        self.cpu_bucket_range = cpu_bucket_range
        self.dram_bucket_range = dram_bucket_range
        self.cpu_indices = cpu_indices
        self.gpu_indices = gpu_indices
        self.pushgateway_url = pushgateway_url
        self.job = job
        self.registry = CollectorRegistry()

        if not gpu_bucket_range:
            raise ValueError(
                "GPU bucket range cannot be empty. Please provide a valid range or omit the argument to use defaults."
            )
        if not cpu_bucket_range:
            raise ValueError(
                "CPU bucket range cannot be empty. Please provide a valid range or omit the argument to use defaults."
            )
        if not dram_bucket_range:
            raise ValueError(
                "DRAM bucket range cannot be empty. Please provide a valid range or omit the argument to use defaults."
            )

        # Initialize GPU histograms
        if self.gpu_indices:
            self.gpu_histograms = Histogram(
                "energy_monitor_gpu_energy_joules",
                "GPU energy consumption",
                ["window", "index"],
                buckets=self.gpu_bucket_range,
                registry=self.registry,
            )
        # Initialize CPU histograms
        if self.cpu_indices:
            self.cpu_histograms = Histogram(
                "energy_monitor_cpu_energy_joules",
                "CPU energy consumption",
                ["window", "index"],
                buckets=self.cpu_bucket_range,
                registry=self.registry,
            )
            # Initialize CPU and DRAM histograms
            if any(cpu.supportsGetDramEnergyConsumption() for cpu in get_cpus().cpus):
                self.dram_histograms = Histogram(
                    "energy_monitor_dram_energy_joules",
                    "DRAM energy consumption",
                    ["window", "index"],
                    buckets=self.dram_bucket_range,
                    registry=self.registry,
                )

        self.max_gpu_bucket = max(self.gpu_bucket_range)
        self.max_cpu_bucket = max(self.cpu_bucket_range)
        self.max_dram_bucket = max(self.dram_bucket_range)

        self.min_gpu_bucket = min(self.gpu_bucket_range)
        self.min_cpu_bucket = min(self.cpu_bucket_range)
        self.min_dram_bucket = min(self.dram_bucket_range)

        self.energy_monitor = ZeusMonitor(
            cpu_indices=cpu_indices, gpu_indices=gpu_indices
        )

    def begin_window(self, name: str, sync_execution: bool = True) -> None:
        """Begin the energy monitoring window.

        Args:
            name (str): The unique name of the measurement window. Must match between calls to 'begin_window' and 'end_window'.
            sync_execution (bool): Whether to execute synchronously. Defaults to True. If assigned True, calls sync_execution_fn with the defined gpu
        """
        if sync_execution:
            sync_execution_fn(self.gpu_indices)

        self.energy_monitor.begin_window(
            f"__EnergyHistogram_{name}", sync_execution=sync_execution
        )

    def end_window(self, name: str, sync_execution: bool = True) -> None:
        """End the current energy monitoring window and record the energy data.

        Retrieves the energy consumption data (for GPUs, CPUs, and DRAMs) for the monitoring window
        and updates the corresponding Histogram metrics. The data is then pushed to the Prometheus Push Gateway.

        Args:
            name (str): The unique name of the measurement window. Must match between calls to 'begin_window' and 'end_window'.
            sync_execution (bool): Whether to execute synchronously. Defaults to True.
        """
        if sync_execution:
            sync_execution_fn(self.gpu_indices)

        measurement = self.energy_monitor.end_window(
            f"__EnergyHistogram_{name}", sync_execution=sync_execution
        )

        if measurement.gpu_energy:
            for gpu_index, gpu_energy in measurement.gpu_energy.items():
                self.gpu_histograms.labels(window=name, index=gpu_index).observe(
                    gpu_energy
                )
                if gpu_energy > self.max_gpu_bucket:
                    warnings.warn(
                        f"GPU {gpu_index} energy {gpu_energy} exceeds the maximum bucket value of {self.max_gpu_bucket}",
                        stacklevel=1,
                    )
                if gpu_energy < self.min_gpu_bucket:
                    warnings.warn(
                        f"GPU {gpu_index} energy {gpu_energy} exceeds the minimum bucket value of {self.min_gpu_bucket}",
                        stacklevel=1,
                    )

        if measurement.cpu_energy:
            for cpu_index, cpu_energy in measurement.cpu_energy.items():
                self.cpu_histograms.labels(window=name, index=cpu_index).observe(
                    cpu_energy
                )
                if cpu_energy > self.max_cpu_bucket:
                    warnings.warn(
                        f"CPU {cpu_index} energy {cpu_energy} exceeds the maximum bucket value of {self.max_cpu_bucket}",
                        stacklevel=1,
                    )
                if cpu_energy < self.min_cpu_bucket:
                    warnings.warn(
                        f"CPU {cpu_index} energy {cpu_energy} exceeds the minimum bucket value of {self.min_cpu_bucket}",
                        stacklevel=1,
                    )

        if measurement.dram_energy:
            for dram_index, dram_energy in measurement.dram_energy.items():
                self.dram_histograms.labels(window=name, index=dram_index).observe(
                    dram_energy
                )
                if dram_energy > self.max_dram_bucket:
                    warnings.warn(
                        f"DRAM {dram_index} energy {dram_energy} exceeds the maximum bucket value of {self.max_dram_bucket}",
                        stacklevel=1,
                    )
                if dram_energy < self.min_dram_bucket:
                    warnings.warn(
                        f"DRAM {dram_index} energy {dram_energy} exceeds the minimum bucket value of {self.min_dram_bucket}",
                        stacklevel=1,
                    )

        push_to_gateway(self.pushgateway_url, job=self.job, registry=self.registry)


class EnergyCumulativeCounter(Metric):
    """EnergyCumulativeCounter class to monitor and record cumulative energy consumption.

    This class tracks GPU, CPU, and DRAM energy usage over time, and records the data as Prometheus Counter metrics.
    The energy consumption metrics are periodically updated and pushed to a Prometheus Push Gateway for monitoring and analysis.

    The cumulative nature of the Counter ensures that energy values are always incremented over time, never reset,
    which is ideal for tracking continuously increasing values like energy usage.
    """

    def __init__(
        self,
        cpu_indices: list,
        gpu_indices: list,
        update_period: int,
        pushgateway_url: str,
        job: str,
    ) -> None:
        """Initialize the EnergyCumulativeCounter.

        Args:
            cpu_indices (list): List of CPU indices to monitor.
            gpu_indices (list): List of GPU indices to monitor.
            update_period: The time interval (in seconds) at which energy measurements are updated.
            pushgateway_url: The URL for the Prometheus Push Gateway where the metrics will be pushed.
            job: The name of the job to be associated with the Prometheus metrics.
        """
        self.cpu_indices = cpu_indices
        self.gpu_indices = gpu_indices
        self.update_period = update_period
        self.pushgateway_url = pushgateway_url
        self.job = job
        self.window_state: dict[str, MonitoringProcessState] = {}

    def begin_window(self, name: str, sync_execution: bool = False) -> None:
        """Begin the energy monitoring window.

        Starts a new multiprocessing process that monitors energy usage periodically
        and pushes the results to the Prometheus Push Gateway.

        Args:
            name (str): The unique name of the measurement window. Must match between calls to 'begin_window' and 'end_window'.
            sync_execution (bool, optional): Whether to execute monitoring synchronously. Defaults to False.
        """
        if sync_execution:
            sync_execution_fn(self.gpu_indices)

        context = mp.get_context("spawn")
        queue = context.Queue()
        proc = context.Process(
            target=energy_monitoring_loop,
            args=(
                name,
                queue,
                self.cpu_indices,
                self.gpu_indices,
                self.update_period,
                self.pushgateway_url,
                self.job,
            ),
        )
        proc.start()
        if not proc.is_alive():
            raise RuntimeError(f"Failed to start monitoring process for {name}.")

        self.window_state[name] = MonitoringProcessState(queue=queue, proc=proc)

    def end_window(self, name: str, sync_execution: bool = False) -> None:
        """End the energy monitoring window.

        Args:
            name (str): The unique name of the measurement window. Must match between calls to 'begin_window' and 'end_window'.
            sync_execution (bool, optional): Whether to execute monitoring synchronously. Defaults to False.
        """
        if name not in self.window_state:
            raise ValueError(f"No active monitoring process found for '{name}'.")

        if sync_execution:
            sync_execution_fn(self.gpu_indices)

        state = self.window_state.pop(name)
        state.queue.put("stop")
        state.proc.join(timeout=20)

        if state.proc.is_alive():
            state.proc.terminate()


def energy_monitoring_loop(
    name: str,
    pipe: mp.Queue,
    cpu_indices: list,
    gpu_indices: list,
    update_period: int,
    pushgateway_url: str,
    job: str,
) -> None:
    """Runs in a separate process to collect and update energy consumption metrics (for GPUs, CPUs, and DRAM).

    Args:
        name (str): The user-defined name of the monitoring window (used as a label for Prometheus metrics).
        pipe (mp.Queue): A multiprocessing queue for inter-process communication, used to signal when to stop the process.
        cpu_indices (list): List of CPU indices to monitor.
        gpu_indices (list): List of GPU indices to monitor.
        update_period (int): The interval (in seconds) between consecutive energy data updates.
        pushgateway_url (str): The URL of the Prometheus Push Gateway where the metrics will be pushed.
        job (str): The name of the Prometheus job associated with these metrics.
    """
    registry = CollectorRegistry()
    energy_monitor = ZeusMonitor(cpu_indices=cpu_indices, gpu_indices=gpu_indices)
    gpu_counters = None
    cpu_counters = None
    dram_counters = None

    if energy_monitor.gpu_indices:
        gpu_counters = Counter(
            "energy_monitor_gpu_energy_joules",
            "GPU energy consumption",
            ["window", "index"],
            registry=registry,
        )

    if energy_monitor.cpu_indices:
        cpu_counters = Counter(
            "energy_monitor_cpu_energy_joules",
            "CPU energy consumption",
            ["window", "index"],
            registry=registry,
        )
        if any(cpu.supportsGetDramEnergyConsumption() for cpu in get_cpus().cpus):
            dram_counters = Counter(
                "energy_monitor_dram_energy_joules",
                "DRAM energy consumption",
                ["window", "index"],
                registry=registry,
            )

    while True:
        if not pipe.empty():
            break
        # Begin and end monitoring window using sync_execution
        energy_monitor.begin_window(
            f"__EnergyCumulativeCounter_{name}", sync_execution=False
        )
        time.sleep(update_period)
        measurement = energy_monitor.end_window(
            f"__EnergyCumulativeCounter_{name}", sync_execution=False
        )

        if measurement.gpu_energy:
            for gpu_index, energy in measurement.gpu_energy.items():
                if gpu_counters:
                    gpu_counters.labels(window=name, index=gpu_index).inc(energy)

        if measurement.cpu_energy:
            for cpu_index, energy in measurement.cpu_energy.items():
                if cpu_counters:
                    cpu_counters.labels(window=name, index=cpu_index).inc(energy)

        if measurement.dram_energy:
            for dram_index, energy in measurement.dram_energy.items():
                if dram_counters:
                    dram_counters.labels(window=name, index=dram_index).inc(energy)
        # Push metrics to Prometheus
        push_to_gateway(pushgateway_url, job=job, registry=registry)


class PowerGauge(Metric):
    """PowerGauge class to monitor and record power consumption.

    This class tracks GPU power usage in real time and records it as **Prometheus Gauge** metrics.
    The Gauge metric type is suitable for tracking values that can go up and down over time, like power consumption.

    Power usage data is collected at regular intervals and pushed to a Prometheus Push Gateway for monitoring.
    """

    def __init__(
        self,
        gpu_indices: list,
        update_period: int,
        pushgateway_url: str,
        job: str,
    ) -> None:
        """Initialize the PowerGauge metric.

        Args:
            gpu_indices (list[int]): List of GPU indices to monitor for power consumption.
            update_period (int): Interval (in seconds) between consecutive power measurements.
            pushgateway_url (str): URL of the Prometheus Push Gateway where Gauge metrics are pushed.
            job (str): Name of the Prometheus job to associate with the power metrics.
        """
        self.gpu_indices = gpu_indices
        self.update_period = update_period
        self.pushgateway_url = pushgateway_url
        self.job = job
        self.window_state: dict[str, MonitoringProcessState] = {}

    def begin_window(self, name: str, sync_execution: bool = False) -> None:
        """Begin the power monitoring window.

        Starts a new multiprocessing process that runs the power monitoring loop.
        The process collects real-time power consumption data and updates the corresponding
        Gauge metrics in Prometheus.

        Args:
            name (str): The unique name of the measurement window. Must match between calls to 'begin_window' and 'end_window'.
            sync_execution (bool, optional): Whether to execute monitoring synchronously. Defaults to False.
        """
        if name in self.window_state:
            raise ValueError(f"PowerGauge metric '{name}' already exists.")

        if sync_execution:
            sync_execution_fn(self.gpu_indices)

        context = mp.get_context("spawn")
        queue = context.Queue()
        proc = context.Process(
            target=power_monitoring_loop,
            args=(
                name,
                queue,
                self.gpu_indices,
                self.update_period,
                self.pushgateway_url,
                self.job,
            ),
        )
        proc.start()
        if not proc.is_alive():
            raise RuntimeError(
                f"Failed to start power monitoring process for '{name}'."
            )

        self.window_state[name] = MonitoringProcessState(queue=queue, proc=proc)

    def end_window(self, name: str, sync_execution: bool = False) -> None:
        """End the power monitoring window.

        Args:
            name (str): The unique name of the measurement window. Must match between calls to 'begin_window' and 'end_window'.
            sync_execution (bool, optional): Whether to execute monitoring synchronously. Defaults to False.
        """
        if sync_execution:
            sync_execution_fn(self.gpu_indices)

        state = self.window_state.pop(name)
        state.queue.put("stop")
        state.proc.join(timeout=20)

        if state.proc.is_alive():
            state.proc.terminate()


def power_monitoring_loop(
    name: str,
    pipe: mp.Queue,
    gpu_indices: list[int],
    update_period: int,
    pushgateway_url: str,
    job: str,
) -> None:
    """Runs in a separate process and periodically collects power consumption data for each GPU and pushes the results to the Prometheus Push Gateway.

    Args:
        name (str): Unique name for the monitoring window (used as a label in Prometheus metrics).
        pipe (multiprocessing.Queue): Queue to receive control signals (e.g., "stop").
        gpu_indices (list[int]): List of GPU indices to monitor for power consumption.
        update_period (int): Interval (in seconds) between consecutive power data polls.
        pushgateway_url (str): URL of the Prometheus Push Gateway where metrics are pushed.
        job (str): Name of the Prometheus job to associate with the metrics.
    """
    power_monitor = PowerMonitor(gpu_indices=gpu_indices)
    registry = CollectorRegistry()

    gpu_gauges = Gauge(
        "power_monitor_gpu_power_watts",
        "Records power consumption for GPU over time",
        ["window", "index"],
        registry=registry,
    )

    while True:
        if not pipe.empty():
            break

        power_measurement = power_monitor.get_power()

        try:
            if power_measurement:
                for gpu_index, power_value in power_measurement.items():
                    gpu_gauges.labels(window=name, index=gpu_index).set(power_value)
        except Exception as e:
            print(f"Error during processing power measurement: {e}")

        try:
            push_to_gateway(pushgateway_url, job=job, registry=registry)
        except Exception as e:
            print(f"Error pushing metrics: {e}")

        time.sleep(update_period)



"""Pydantic models for Batch size/Trials/GaussianTsArmState."""

from __future__ import annotations

from datetime import datetime
from typing import Any, Optional

from zeus.utils.pydantic_v1 import Field, root_validator, validator, BaseModel
from zeus.optimizer.batch_size.server.database.schema import (
    GaussianTsArmStateTable,
    TrialStatus,
    TrialType,
)


class BatchSizeBase(BaseModel):
    """Base model for representing batch size.

    Attributes:
        job_id (str): The ID of the job.
        batch_size (int): The size of the batch (greater than 0).
    """

    job_id: str
    batch_size: int = Field(gt=0)

    class Config:  # type: ignore
        """Model configuration.

        Make it immutable after it's created.
        """

        frozen = True


class Trial(BatchSizeBase):
    """Pydantic model that represents Trial.

    Attributes:
        job_id (str): The ID of the job.
        batch_size (int): The size of the batch (greater than 0).
        trial_number (int): Number of trial.
        start_timestamp (datetime): Start time of trial.
        end_timestamp (datetime): End time of trial.
        type (TrialType): Type of this trial, which means in which stage this trial was executed.
        status (TrialStatus): Status of trial
        time (Optional[float]): Total time consumption of this trial.
        energy (Optional[float]): Total energy consumption of this trial.
        converged (Optional[bool]): Whether this trial is converged or not.
    """

    trial_number: int = Field(gt=0)
    start_timestamp: datetime
    end_timestamp: Optional[datetime] = Field(None)
    type: TrialType
    status: TrialStatus
    time: Optional[float] = Field(None, ge=0)
    energy: Optional[float] = Field(None, ge=0)
    converged: Optional[bool] = None

    class Config:  # type: ignore
        """Model configuration.

        Enable instantiating model from an ORM object, and make it immutable after it's created.
        """

        orm_mode = True
        frozen = True

    @root_validator(skip_on_failure=True)
    def _validate_mab(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate trial.

        We are checking
            - start_timestamp <= end_timestamp
            - if status == dispatched | Failed, time/energy/converged = None
                else time/energy/converged != None
        """
        start_timestamp: datetime = values["start_timestamp"]
        end_timestamp: datetime | None = values["end_timestamp"]
        status: TrialStatus = values["status"]
        time: float | None = values["time"]
        energy: float | None = values["energy"]
        converged: bool | None = values["converged"]

        if end_timestamp is not None and start_timestamp > end_timestamp:
            raise ValueError(
                f"start is earlier than end: {start_timestamp} > {end_timestamp}"
            )
        if status in (TrialStatus.Dispatched, TrialStatus.Failed):
            if not (time is None and energy is None and converged is None):
                raise ValueError("Trial status and result is not matching.")
            if status == TrialStatus.Failed and end_timestamp is None:
                raise ValueError("Trial ended but end_timestamp is None.")
        elif (
            time is None or energy is None or converged is None or end_timestamp is None
        ):
            raise ValueError(
                f"Trial ended but the result is incomplete: time({time}), energy({energy}), converged({converged}), end_timestamp({end_timestamp})"
            )

        return values


class GaussianTsArmState(BatchSizeBase):
    """Model representing Gaussian Thompson Sampling arm state.

    Attributes:
        param_mean (float): Mean of the belief prior distribution.
        param_precision (float): Precision of the belief prior distribution.
        reward_precision (float): Precision (inverse variance) of the reward distribution.
        num_observations (int): How many observations we made.
    """

    param_mean: float
    param_precision: float
    reward_precision: float
    num_observations: int = Field(ge=0)

    class Config:  # type: ignore
        """Model configuration.

        Enable instantiating model from an ORM object, and make it immutable after it's created.
        """

        orm_mode = True
        frozen = True

    def to_orm(self) -> GaussianTsArmStateTable:
        """Convert pydantic model to ORM object.

        Returns:
            GaussianTsArmState: The ORM object of Gaussian Arm State.
        """
        d = self.dict()
        g = GaussianTsArmStateTable()
        for k, v in d.items():
            setattr(g, k, v)
        return g


# Helper models


class TrialResult(BatchSizeBase):
    """Model for reading the result of the trial.

    Refer to [`Trial`][zeus.optimizer.batch_size.server.batch_size_state.models.Trial] for attributes.
    """

    trial_number: int = Field(gt=0)
    status: TrialStatus
    time: float = Field(ge=0)
    energy: float = Field(ge=0)
    converged: bool

    class Config:  # type: ignore
        """Model configuration.

        Enable instantiating model from an ORM object, and make it immutable after it's created.
        """

        orm_mode = True
        frozen = True

    @validator("status")
    def _check_state(cls, s: TrialStatus) -> TrialStatus:
        """Check if status is equal to succeeded."""
        if s == TrialStatus.Succeeded:
            return s
        else:
            raise ValueError(f"{s} should be succeeded to have a valid result.")


class TrialResultsPerBs(BatchSizeBase):
    """Model representing all succeeded results of trial for a given batch size.

    Attributes:
        results (list[TrialResult]): List of TrialResult per batch size.
    """

    results: list[TrialResult]

    @root_validator(skip_on_failure=True)
    def _check_explorations(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate if job_id and bs are consistent across all items in results."""
        bs: int = values["batch_size"]
        job_id: str = values["job_id"]
        ms: list[TrialResult] = values["results"]
        ms.sort(key=lambda x: x.trial_number, reverse=True)

        for m in ms:
            if job_id != m.job_id:
                raise ValueError(
                    f"job_id doesn't correspond with results: {job_id} != {m.job_id}"
                )
            if bs != m.batch_size:
                raise ValueError(
                    f"Batch size doesn't correspond with results: {bs} != {m.batch_size}"
                )
            if m.status != TrialStatus.Succeeded:
                raise ValueError(
                    f"This list should only contain succeeded trials. Encounted trial({m.trial_number}) of status = {m.status}"
                )

        return values


class ExplorationsPerJob(BaseModel):
    """Model representing all succeeded explorations we have done for a job. Immutable after it's created.

    Attributes:
        job_id (str): The ID of the job.
        explorations_per_bs (dict[int, list[Trial]]): Dictionary of "succeeded" explorations per batch size in trial_number ascending order.
    """

    job_id: str
    explorations_per_bs: dict[int, list[Trial]]  # BS -> Trials with exploration type

    class Config:  # type: ignore
        """Model configuration.

        Make it immutable after it's created.
        """

        frozen = True

    @root_validator(skip_on_failure=True)
    def _check_explorations(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Check bs and job_id corresponds to explorations_per_bs and batch size is consistent."""
        job_id: str = values["job_id"]
        exps_per_bs: dict[int, list[Trial]] = values["explorations_per_bs"]

        for bs, exps in exps_per_bs.items():
            # Sort ascending just in case. Sql will return asc order anyways.
            exps.sort(key=lambda x: x.trial_number)
            for exp in exps:
                if job_id != exp.job_id:
                    raise ValueError(
                        f"job_id doesn't correspond with explorations: {job_id} != {exp.job_id}"
                    )
                if bs != exp.batch_size:
                    raise ValueError(
                        f"Batch size doesn't correspond with explorations: {bs} != {exp.batch_size}"
                    )
                if exp.type != TrialType.Exploration:
                    raise ValueError("Trial type is not equal to Exploration.")
                if exp.status == TrialStatus.Failed:
                    raise ValueError("Should not include failed trial.")

        return values



"""Test metric.py."""

from __future__ import annotations

from unittest.mock import patch, MagicMock

import pytest

from zeus.metric import EnergyHistogram, EnergyCumulativeCounter, PowerGauge


@pytest.fixture
def mock_get_cpus():
    """Fixture to mock `get_cpus()` to avoid RAPL-related errors."""
    with patch("zeus.metric.get_cpus", autospec=True) as mock_get_cpus:
        mock_cpu_0 = MagicMock()
        mock_cpu_0.supportsGetDramEnergyConsumption.return_value = True
        mock_cpu_1 = MagicMock()
        mock_cpu_1.supportsGetDramEnergyConsumption.return_value = False

        mock_get_cpus.return_value.cpus = [mock_cpu_0, mock_cpu_1]
        yield mock_get_cpus


@pytest.fixture
def mock_zeus_monitor():
    """Fixture to mock ZeusMonitor behavior."""
    with patch("zeus.metric.ZeusMonitor", autospec=True) as zeus_monitor:
        mock_instance = zeus_monitor.return_value
        mock_instance.end_window.return_value = MagicMock(
            gpu_energy={0: 50.0, 1: 100.0, 2: 200.0},
            cpu_energy={0: 40.0, 1: 50.0},
            dram_energy={0: 10},
        )
        mock_instance.gpu_indices = [0, 1, 2]
        mock_instance.cpu_indices = [0, 1]
        yield mock_instance


@pytest.fixture
def mock_power_monitor():
    """Fixture to mock PowerMonitor."""
    with patch("zeus.metric.PowerMonitor", autospec=True) as power_monitor:
        mock_instance = power_monitor.return_value
        mock_instance.get_power.return_value = {
            0: 300.0,
            1: 310.0,
            2: 320.0,
        }
        yield mock_instance


@pytest.fixture
def mock_histogram():
    """Fixture to mock Prometheus Histogram creation.

    Mocks the Histogram functionality to avoid real Prometheus interactions
    and to validate histogram-related method calls.
    """
    with patch("zeus.metric.Histogram", autospec=True) as histogram:
        yield histogram


def test_energy_histogram(
    mock_get_cpus: MagicMock,
    mock_zeus_monitor: MagicMock,
    mock_histogram: MagicMock,
) -> None:
    """Test EnergyHistogram class."""
    cpu_indices = [0, 1]
    gpu_indices = [0, 1, 2]
    pushgateway_url = "http://localhost:9091"
    window_name = "test_window"

    # Ensure mocked CPUs have the required method
    mock_get_cpus.return_value.cpus = [
        MagicMock(supportsGetDramEnergyConsumption=MagicMock(return_value=True)),
        MagicMock(supportsGetDramEnergyConsumption=MagicMock(return_value=False)),
    ]

    histogram_metric = EnergyHistogram(
        cpu_indices=cpu_indices,
        gpu_indices=gpu_indices,
        pushgateway_url=pushgateway_url,
        job="test_energy_histogram",
    )

    # Mock single Histogram objects for GPU, CPU, and DRAM
    gpu_mock_histogram = mock_histogram(
        name="gpu_energy_histogram",
        documentation="Mocked GPU histogram",
        labelnames=["window", "index"],
    )

    cpu_mock_histogram = mock_histogram(
        name="cpu_energy_histogram",
        documentation="Mocked CPU histogram",
        labelnames=["window", "index"],
    )

    dram_mock_histogram = mock_histogram(
        name="dram_energy_histogram",
        documentation="Mocked DRAM histogram",
        labelnames=["window", "index"],
    )

    # Attach mocked histograms to the metric
    histogram_metric.gpu_histograms = gpu_mock_histogram
    histogram_metric.cpu_histograms = cpu_mock_histogram
    histogram_metric.dram_histograms = dram_mock_histogram

    # Begin and end the monitoring window
    histogram_metric.begin_window(window_name, False)
    with patch("http.client.HTTPConnection", autospec=True) as mock_http:
        mock_http_instance = mock_http.return_value
        mock_http_instance.getresponse.return_value.code = 200
        mock_http_instance.getresponse.return_value.msg = "OK"
        mock_http_instance.getresponse.return_value.info = lambda: {}
        mock_http_instance.sock = MagicMock()
        histogram_metric.end_window(window_name, False)

    # Validate GPU histogram observations
    for (
        gpu_index,
        energy,
    ) in mock_zeus_monitor.return_value.end_window.return_value.gpu_energy.items():
        gpu_mock_histogram.labels.assert_any_call(window=window_name, index=gpu_index)
        gpu_mock_histogram.labels.return_value.observe.assert_any_call(energy)

    # Validate CPU histogram observations
    for (
        cpu_index,
        energy,
    ) in mock_zeus_monitor.return_value.end_window.return_value.cpu_energy.items():
        cpu_mock_histogram.labels.assert_any_call(window=window_name, index=cpu_index)
        cpu_mock_histogram.labels.return_value.observe.assert_any_call(energy)

    # Validate DRAM histogram observations
    for (
        dram_index,
        energy,
    ) in mock_zeus_monitor.return_value.end_window.return_value.dram_energy.items():
        dram_mock_histogram.labels.assert_any_call(window=window_name, index=dram_index)
        dram_mock_histogram.labels.return_value.observe.assert_any_call(energy)


@patch("zeus.metric.energy_monitoring_loop", autospec=True)
@patch("zeus.metric.mp.get_context", autospec=True)
def test_energy_cumulative_counter(
    mock_mp_context: MagicMock,
    mock_energy_monitoring_loop: MagicMock,
):
    """Test EnergyCumulativeCounter with mocked subprocess behavior."""
    cpu_indices = [0, 1]
    gpu_indices = [0, 1, 2]
    pushgateway_url = "http://localhost:9091"

    # Mock the context and queue
    mock_queue = MagicMock()
    mock_process = MagicMock()
    mock_mp_context.return_value.Queue.return_value = mock_queue
    mock_mp_context.return_value.Process.return_value = (
        mock_process  # Ensure Process returns mock_process
    )

    # Mock the behavior of subprocess
    mock_energy_monitoring_loop.return_value = (
        None  # Simulate the subprocess running without errors
    )

    # Create the EnergyCumulativeCounter instance
    cumulative_counter = EnergyCumulativeCounter(
        cpu_indices=cpu_indices,
        gpu_indices=gpu_indices,
        update_period=2,
        pushgateway_url=pushgateway_url,
        job="test_energy_counter",
    )

    # Begin a monitoring window
    cumulative_counter.begin_window("test_counter")

    # Assert that the subprocess was started with the correct arguments
    mock_mp_context.return_value.Process.assert_called_once_with(
        target=mock_energy_monitoring_loop,
        args=(
            "test_counter",
            mock_queue,
            cpu_indices,
            gpu_indices,
            2,
            pushgateway_url,
            "test_energy_counter",
        ),
    )
    mock_process.start.assert_called_once()

    # Assert the window state is updated correctly
    assert "test_counter" in cumulative_counter.window_state
    state = cumulative_counter.window_state["test_counter"]
    assert state.queue == mock_queue
    assert state.proc == mock_process

    # End the monitoring window
    cumulative_counter.end_window("test_counter")
    mock_queue.put.assert_called_once_with("stop")
    mock_process.join.assert_called_once()
    assert "test_counter" not in cumulative_counter.window_state


@patch("zeus.metric.power_monitoring_loop", autospec=True)
@patch("zeus.metric.mp.get_context", autospec=True)
def test_power_gauge(
    mock_mp_context: MagicMock,
    mock_power_monitoring_loop: MagicMock,
):
    """Test PowerGauge with mocked subprocess behavior."""
    gpu_indices = [0, 1, 2]
    pushgateway_url = "http://localhost:9091"

    # Mock the context and queue
    mock_queue = MagicMock()
    mock_process = MagicMock()
    mock_mp_context.return_value.Queue.return_value = mock_queue
    mock_mp_context.return_value.Process.return_value = mock_process

    # Mock the behavior of subprocess
    mock_power_monitoring_loop.return_value = (
        None  # Simulate the subprocess running without errors
    )

    # Create the EnergyCumulativeCounter instance
    power_gauge = PowerGauge(
        gpu_indices=gpu_indices,
        update_period=2,
        pushgateway_url=pushgateway_url,
        job="test_power_gauge",
    )

    # Begin a monitoring window
    power_gauge.begin_window("test_power_gauge")

    # Assert that the subprocess was started with the correct arguments
    mock_mp_context.return_value.Process.assert_called_once_with(
        target=mock_power_monitoring_loop,
        args=(
            "test_power_gauge",
            mock_queue,
            gpu_indices,
            2,
            pushgateway_url,
            "test_power_gauge",
        ),
    )
    mock_process.start.assert_called_once()

    # Assert the window state is updated correctly
    assert "test_power_gauge" in power_gauge.window_state
    state = power_gauge.window_state["test_power_gauge"]
    assert state.queue == mock_queue
    assert state.proc == mock_process

    # End the monitoring window
    power_gauge.end_window("test_power_gauge")
    mock_queue.put.assert_called_once_with("stop")
    mock_process.join.assert_called_once()
    assert "test_power_gauge" not in power_gauge.window_state



"""NVIDIA Jetson platform support."""

from __future__ import annotations

import abc
import asyncio
import atexit
import enum
import os
import platform
import sys
import time
import multiprocessing as mp
from dataclasses import dataclass, asdict, fields
from pathlib import Path
from queue import Empty
from typing import TypedDict

from zeus.device.soc.common import SoC, SoCMeasurement, ZeusSoCInitError


class ZeusJetsonInitError(ZeusSoCInitError):
    """Jetson initialization failures."""

    def __init__(self, message: str) -> None:
        """Initialize Zeus Exception."""
        super().__init__(message)


def check_file(path: Path) -> bool:
    """Check if the given path exists and is a file."""
    return path.exists() and path.is_file()


class PowerMeasurementStrategy(abc.ABC):
    """Abstract base class for two different power measurement strategies."""

    @abc.abstractmethod
    def measure_power(self) -> float:
        """Measure power in mW."""
        pass


class DirectPower(PowerMeasurementStrategy):
    """Reads power directly from a sysfs path."""

    def __init__(self, power_path: Path) -> None:
        """Initialize DirectPower paths."""
        self.power_path: Path = power_path

    def measure_power(self) -> float:
        """Measure power by reading from sysfs paths.

        Units: mW.
        """
        power: float = float(self.power_path.read_text().strip())
        return power


class VoltageCurrentProduct(PowerMeasurementStrategy):
    """Computes power as product of voltage and current, read from two sysfs paths."""

    def __init__(self, voltage_path: Path, current_path: Path) -> None:
        """Initialize VoltageCurrentProduct paths."""
        self.voltage_path: Path = voltage_path
        self.current_path: Path = current_path

    def measure_power(self) -> float:
        """Measure power by reading from sysfs paths.

        Units: mW.
        """
        voltage: float = float(self.voltage_path.read_text().strip())
        current: float = float(self.current_path.read_text().strip())
        return (voltage * current) / 1000


@dataclass
class JetsonMeasurement(SoCMeasurement):
    """Represents energy measurements for Jetson subsystems.

    All measurements are in mJ.
    """

    cpu_energy_mj: float | None = None
    gpu_energy_mj: float | None = None
    total_energy_mj: float | None = None

    def __sub__(self, other: JetsonMeasurement) -> JetsonMeasurement:
        """Produce a single measurement object containing differences across all fields."""
        if not isinstance(other, type(self)):
            raise TypeError("Subtraction is only supported between Jetson instances.")

        result = self.__class__()

        for field in fields(self):
            f_name = field.name
            value1 = getattr(self, f_name)
            value2 = getattr(other, f_name)
            if value1 is None and value2 is None:
                continue
            else:
                setattr(result, f_name, value1 - value2)

        return result

    def zeroAllFields(self) -> None:
        """Set all internal measurement values to zero."""
        for field in fields(self):
            f_name = field.name
            f_value = getattr(self, f_name)
            if isinstance(f_value, float):
                setattr(self, f_name, 0.0)
            else:
                setattr(self, f_name, None)


class DeviceMap(TypedDict, total=False):
    """Map of device names to their corresponding power measurement strategies."""

    cpu_power_mw: PowerMeasurementStrategy
    gpu_power_mw: PowerMeasurementStrategy
    total_power_mw: PowerMeasurementStrategy


class Jetson(SoC):
    """An interface for obtaining the energy metrics of a Jetson processor."""

    def __init__(self) -> None:
        """Initialize an instance of a Jetson energy monitor."""
        if not jetson_is_available():
            raise ZeusJetsonInitError(
                "No Jetson processor was detected on the current device."
            )

        super().__init__()

        # Maps each power rail (cpu, gpu, and total) to a power measurement strategy
        self.power_measurement = self._discover_available_metrics()
        self.available_metrics: set[str] | None = None

        # Spawn polling process
        context = mp.get_context("spawn")
        self.command_queue = context.Queue()
        self.result_queue = context.Queue()
        self.process = context.Process(
            target=_polling_process_async_wrapper,
            args=(self.command_queue, self.result_queue, self.power_measurement),
        )
        self.process.start()
        atexit.register(self._stop_process)

    def _discover_available_metrics(self) -> DeviceMap:
        """Return available power measurement metrics per rail from the INA3221 sensor on Jetson devices.

        All official NVIDIA Jetson devices have at least 1 INA3221 power monitor that measures per-rail power usage via 3 channels.

          - https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-3276/index.html#page/Tegra%20Linux%20Driver%20Package%20Development%20Guide/clock_power_setup.html#
          - https://docs.nvidia.com/jetson/archives/r35.6.1/DeveloperGuide/SD/PlatformPowerAndPerformance/JetsonXavierNxSeriesAndJetsonAgxXavierSeries.html#software-based-power-consumption-modeling
          - https://docs.nvidia.com/jetson/archives/r36.4.3/DeveloperGuide/SD/PlatformPowerAndPerformance/JetsonOrinNanoSeriesJetsonOrinNxSeriesAndJetsonAgxOrinSeries.html#
        """
        path = Path("/sys/bus/i2c/drivers/ina3221x")

        metric_paths: dict[str, dict[str, Path]] = {}
        power_measurement: DeviceMap = {}

        def extract_directories(
            path: Path, rail_name: str, rail_index: str, type: str
        ) -> None:
            """Extract file paths for power, voltage, and current measurements based on the rail naming type."""
            rail_name_lower = rail_name.lower()

            if "cpu" in rail_name_lower:
                rail_name_simplified = "cpu_power_mw"
            elif "gpu" in rail_name_lower:
                rail_name_simplified = "gpu_power_mw"
            elif (
                "system" in rail_name_lower
                or "_in" in rail_name_lower
                or "total" in rail_name_lower
            ):
                rail_name_simplified = "total_power_mw"
            else:
                return  # Skip unsupported rail types

            if type == "label":
                power_path = path / f"power{rail_index}_input"
                volt_path = path / f"in{rail_index}_input"
                curr_path = path / f"curr{rail_index}_input"
            else:
                power_path = path / f"in_power{rail_index}_input"
                volt_path = path / f"in_voltage{rail_index}_input"
                curr_path = path / f"in_current{rail_index}_input"

            if check_file(power_path):
                metric_paths[rail_name_simplified] = {"power": Path(power_path)}
            elif check_file(volt_path) and check_file(curr_path):
                metric_paths[rail_name_simplified] = {
                    "volt": Path(volt_path),
                    "curr": Path(curr_path),
                }
            # Else, skip the rail due to insufficient metrics for power

        for device in path.glob("*"):
            for subdevice in device.glob("*"):
                # Get the files containing rail names.
                label_files = subdevice.glob("in*_label")
                rail_files = subdevice.glob("rail_name_*")
                # For each rail name, get its respective power, voltage, current paths.
                for label_file in label_files:
                    rail_name = label_file.read_text().strip()
                    rail_index = label_file.name.split("_")[0].lstrip("in")
                    extract_directories(subdevice, rail_name, rail_index, "label")
                for rail_file in rail_files:
                    rail_name = rail_file.read_text().strip()
                    rail_index = rail_file.name.split("rail_name_", 1)[-1]
                    extract_directories(subdevice, rail_name, rail_index, "rail_name")

        # Instantiate PowerMeasurementStrategy objects based on available metrics
        for rail, metrics in metric_paths.items():
            if "power" in metrics:
                power_measurement[rail] = DirectPower(metrics["power"])
            elif "volt" in metrics and "curr" in metrics:
                power_measurement[rail] = VoltageCurrentProduct(
                    metrics["volt"], metrics["curr"]
                )
            # Else, skip the rail due to insufficient metrics for power
        return power_measurement

    def getAvailableMetrics(self) -> set[str]:
        """Return a set of all observable metrics on the Jetson device."""
        if self.available_metrics is None:
            result: JetsonMeasurement = self.getTotalEnergyConsumption()
            available_metrics = set()

            metrics_dict = asdict(result)
            for f_name, f_value in metrics_dict.items():
                if f_value is not None:
                    available_metrics.add(f_name)

            self.available_metrics = available_metrics
        return self.available_metrics

    def _stop_process(self) -> None:
        """Kill the polling process."""
        self.command_queue.put_nowait(Command.STOP)
        self.process.join(timeout=1.0)
        self.process.kill()

    def getTotalEnergyConsumption(self, timeout: float = 15.0) -> JetsonMeasurement:
        """Returns the total energy consumption of the Jetson device. This measurement is cumulative.

        Units: mJ.
        """
        self.command_queue.put(Command.READ)
        return self.result_queue.get(timeout=timeout)


class Command(enum.Enum):
    """Provide commands for the polling process."""

    READ = "read"
    STOP = "stop"


def _polling_process_async_wrapper(
    command_queue: mp.Queue[Command],
    result_queue: mp.Queue[JetsonMeasurement],
    power_measurement: DeviceMap,
) -> None:
    """Function wrapper for the asynchronous energy polling process."""
    asyncio.run(
        _polling_process_async(
            command_queue,
            result_queue,
            power_measurement,
        )
    )


async def _polling_process_async(
    command_queue: mp.Queue[Command],
    result_queue: mp.Queue[JetsonMeasurement],
    power_measurement: DeviceMap,
) -> None:
    """Continuously polls for accumulated energy measurements for CPU, GPU, and total power, listening for commands to stop or return the measurement."""
    cumulative_measurement = JetsonMeasurement(
        cpu_energy_mj=0.0 if "cpu_power_mw" in power_measurement else None,
        gpu_energy_mj=0.0 if "gpu_power_mw" in power_measurement else None,
        total_energy_mj=0.0 if "total_power_mw" in power_measurement else None,
    )

    prev_ts = time.monotonic()

    while True:
        current_ts: float = time.monotonic()
        dt: float = current_ts - prev_ts

        if "cpu_power_mw" in power_measurement:
            cpu_power_mw = power_measurement["cpu_power_mw"].measure_power()
            cpu_energy_mj = cpu_power_mw * dt
            cumulative_measurement.cpu_energy_mj = (
                cumulative_measurement.cpu_energy_mj or 0.0
            ) + cpu_energy_mj
        if "gpu_power_mw" in power_measurement:
            gpu_power_mw = power_measurement["gpu_power_mw"].measure_power()
            gpu_energy_mj = gpu_power_mw * dt
            cumulative_measurement.gpu_energy_mj = (
                cumulative_measurement.gpu_energy_mj or 0.0
            ) + gpu_energy_mj
        if "total_power_mw" in power_measurement:
            total_power_mw = power_measurement["total_power_mw"].measure_power()
            total_energy_mj = total_power_mw * dt
            cumulative_measurement.total_energy_mj = (
                cumulative_measurement.total_energy_mj or 0.0
            ) + total_energy_mj

        prev_ts = current_ts

        try:
            command = await asyncio.to_thread(
                command_queue.get,
                timeout=0.1,
            )
        except Empty:
            # Update energy and do nothing
            continue

        if command == Command.STOP:
            break
        if command == Command.READ:
            # Update and return energy measurement
            result_queue.put(cumulative_measurement)


def jetson_is_available() -> bool:
    """Return if the current processor is a Jetson device."""
    if sys.platform != "linux" or platform.processor() != "aarch64":
        return False

    return os.path.exists("/usr/lib/aarch64-linux-gnu/tegra") or os.path.exists(
        "/etc/nv_tegra_release"
    )



"""Abstraction layer for GPU devices.

The main function of this module is [`get_gpus`][zeus.device.gpu.get_gpus],
which returns a GPU Manager object specific to the platform.

!!! Important
    In theory, any NVIDIA GPU would be supported.
    On the other hand, for AMD GPUs, we currently only support ROCm 6.1 and later.

## Getting handles to GPUs

The main API exported from this module is the `get_gpus` function. It returns either
[`NVIDIAGPUs`][zeus.device.gpu.nvidia.NVIDIAGPUs] or [`AMDGPUs`][zeus.device.gpu.amd.AMDGPUs]
depending on the platform.

```python
from zeus.device import get_gpus
gpus = get_gpus()
```

## Calling GPU management APIs

GPU management library APIs are mapped to methods on [`GPU`][zeus.device.gpu.common.GPU].

For example, for NVIDIA GPUs (which uses `pynvml`), you would have called:

```python
handle = pynvml.nvmlDeviceGetHandleByIndex(gpu_index)
constraints = pynvml.nvmlDeviceGetPowerManagementLimitConstraints(handle)
```

With the Zeus GPU abstraction layer, you would now call:

```python
gpus = get_gpus() # returns an NVIDIAGPUs object
constraints = gpus.getPowerManagementLimitConstraints(gpu_index)
```

## Non-blocking calls

Some implementations of `GPU` support non-blocking calls to setters.
If non-blocking calls are not supported, setting `block` will be ignored and the call will block.
Check [`GPU.supports_non_blocking`][zeus.device.gpu.common.GPU.supports_nonblocking_setters]
to see if non-blocking calls are supported.
Note that non-blocking calls will not raise exceptions even if the call fails.

Currently, only [`ZeusdNVIDIAGPU`][zeus.device.gpu.nvidia.ZeusdNVIDIAGPU] supports non-blocking calls
to methods that set the GPU's power limit, GPU frequency, memory frequency, and persistence mode.
This is possible because the Zeus daemon supports a `block: bool` parameter in HTTP requests,
which can be set to `False` to make the call return immediately without checking the result.

## Error handling

The following exceptions are defined in this module:

- [`ZeusGPUInitError`][zeus.device.gpu.ZeusGPUInitError]: Base class for initialization errors.
- [`ZeusGPUInvalidArgError`][zeus.device.gpu.ZeusGPUInvalidArgError]: Error for invalid arguments.
- [`ZeusGPUNotSupportedError`][zeus.device.gpu.ZeusGPUNotSupportedError]: Error for unsupported GPUs.
- [`ZeusGPUNoPermissionError`][zeus.device.gpu.ZeusGPUNoPermissionError]: Error for permission issues.
- [`ZeusGPUAlreadyInitializedError`][zeus.device.gpu.ZeusGPUAlreadyInitializedError]: Error for reinitialization.
- [`ZeusGPUNotFoundError`][zeus.device.gpu.ZeusGPUNotFoundError]: Error for missing GPUs.
- [`ZeusGPUInsufficientSizeError`][zeus.device.gpu.ZeusGPUInsufficientSizeError]: Error for insufficient buffer size.
- [`ZeusGPUInsufficientPowerError`][zeus.device.gpu.ZeusGPUInsufficientPowerError]: Error for insufficient power.
- [`ZeusGPUDriverNotLoadedError`][zeus.device.gpu.ZeusGPUDriverNotLoadedError]: Error for driver issues.
- [`ZeusGPUTimeoutError`][zeus.device.gpu.ZeusGPUTimeoutError]: Error for timeout issues.
- [`ZeusGPUIRQError`][zeus.device.gpu.ZeusGPUIRQError]: Error for IRQ issues.
- [`ZeusGPULibraryNotFoundError`][zeus.device.gpu.ZeusGPULibraryNotFoundError]: Error for missing libraries.
- [`ZeusGPUFunctionNotFoundError`][zeus.device.gpu.ZeusGPUFunctionNotFoundError]: Error for missing functions.
- [`ZeusGPUCorruptedInfoROMError`][zeus.device.gpu.ZeusGPUCorruptedInfoROMError]: Error for corrupted info ROM.
- [`ZeusGPULostError`][zeus.device.gpu.ZeusGPULostError]: Error for lost GPUs.
- [`ZeusGPUResetRequiredError`][zeus.device.gpu.ZeusGPUResetRequiredError]: Error for GPUs requiring reset.
- [`ZeusGPUOperatingSystemError`][zeus.device.gpu.ZeusGPUOperatingSystemError]: Error for OS issues.
- [`ZeusGPULibRMVersionMismatchError`][zeus.device.gpu.ZeusGPULibRMVersionMismatchError]: Error for library version mismatch.
- [`ZeusGPUMemoryError`][zeus.device.gpu.ZeusGPUMemoryError]: Error for memory issues.
- [`ZeusGPUUnknownError`][zeus.device.gpu.ZeusGPUUnknownError]: Error for unknown issues.
"""

from __future__ import annotations

from zeus.device.gpu.common import *
from zeus.device.gpu.common import GPUs, ZeusGPUInitError
from zeus.device.gpu.nvidia import nvml_is_available, NVIDIAGPUs
from zeus.device.gpu.amd import amdsmi_is_available, AMDGPUs


_gpus: GPUs | None = None


def get_gpus(ensure_homogeneous: bool = False) -> GPUs:
    """Initialize and return a singleton object for GPU management.

    This function returns a GPU management object that aims to abstract
    the underlying GPU vendor and their specific monitoring library
    (pynvml for NVIDIA GPUs and amdsmi for AMD GPUs). Management APIs
    are mapped to methods on the returned [`GPUs`][zeus.device.gpu.GPUs] object.

    GPU availability is checked in the following order:

    1. NVIDIA GPUs using `pynvml`
    1. AMD GPUs using `amdsmi`
    1. If both are unavailable, a `ZeusGPUInitError` is raised.

    Args:
        ensure_homogeneous (bool): If True, ensures that all tracked GPUs have the same name.
    """
    global _gpus
    if _gpus is not None:
        return _gpus

    if nvml_is_available():
        _gpus = NVIDIAGPUs(ensure_homogeneous)
        return _gpus
    elif amdsmi_is_available():
        _gpus = AMDGPUs(ensure_homogeneous)
        return _gpus
    else:
        raise ZeusGPUInitError(
            "NVML and AMDSMI unavailable. Failed to initialize GPU management library."
        )



"""The JobManager singleton class manages all job states."""

from __future__ import annotations

import time
import asyncio
import traceback

from fastapi import HTTPException

from zeus.optimizer.pipeline_frequency.common import (
    JobInfo,
    PFOServerSettings,
    FrequencySchedule,
    ProfilingResult,
    RankInfo,
    save_prof,
    save_sched,
    save_ranks,
)
from zeus.utils.logging import get_logger
from zeus.utils.async_utils import create_task

GLOBAL_JOB_MANAGER: JobManager | None = None

logger = get_logger(__name__)


class JobManager:
    """A singleton class that manages all states."""

    def __init__(self, pfo_settings: PFOServerSettings) -> None:
        """Initialize the job manager."""
        self.pfo_settings = pfo_settings

        self._job_infos: dict[str, JobInfo] = {}
        self._job_rank_infos: dict[str, list[RankInfo]] = {}
        self._job_tasks: dict[str, asyncio.Task] = {}
        self._job_result_channels: dict[str, asyncio.Queue[ProfilingResult]] = {}
        self._job_sched_request_channels: dict[str, asyncio.Queue] = {}
        self._job_sched_response_channels: dict[str, list[asyncio.Queue]] = {}
        self._job_last_active_time: dict[str, float] = {}

        # Spawn cleanup task that evicts the state of jobs that have not been active
        # for a long time.
        create_task(
            self._cleanup_task(
                cleanup_period=60,
                max_idle_time=pfo_settings.max_job_idle_time,
            ),
            logger=logger,
        )

    def register_job(self, job_info: JobInfo) -> None:
        """Prepare internal state for a new job.

        This method will be invoked exactly once by the global rank 0 (master) process.
        """
        job_id = job_info.job_id
        world_size = job_info.world_size
        self._job_infos[job_id] = job_info
        self._job_rank_infos[job_id] = []
        self._job_result_channels[job_id] = asyncio.Queue(maxsize=world_size)
        self._job_sched_request_channels[job_id] = asyncio.Queue(maxsize=world_size)
        self._job_sched_response_channels[job_id] = [
            asyncio.Queue(maxsize=1) for _ in range(world_size)
        ]
        self._job_tasks[job_id] = create_task(
            self._job_task(job_id, self.pfo_settings.dump_data),
            logger=logger,
        )
        self._job_last_active_time[job_id] = time.monotonic()

    def register_rank(self, job_id: str, rank_info: RankInfo) -> None:
        """Register rank-specific information for an already registered job.

        This method will be invoked `world_size` number of times (once per rank).
        """
        self._job_rank_infos[job_id].append(rank_info)
        self._job_last_active_time[job_id] = time.monotonic()

    async def get_frequency_schedule(self, job_id: str, rank: int) -> FrequencySchedule:
        """Get the next frequency schedule for a rank.

        This method will be called `world_size` number of times (once per rank).
        All ranks will block on this method untill everyone reports their
        profiling results and calls this method.

        When an internal scheduler error happened at any point of servicing the
        job, clients will be notified through this API with a 500 Internal Error.
        """
        await self._job_sched_request_channels[job_id].put(rank)
        res = await self._job_sched_response_channels[job_id][rank].get()
        if isinstance(res, Exception):
            code = 400 if isinstance(res, ValueError) else 500
            raise HTTPException(
                status_code=code,
                detail="".join(
                    traceback.format_exception(type(res), res, res.__traceback__)
                ),
            )
        self._job_last_active_time[job_id] = time.monotonic()
        return res

    def report_profiling_result(self, job_id: str, result: ProfilingResult) -> None:
        """Send the profiling result to the job task and immediately return.

        This method will be called `world_size` number of times - one for each rank.
        """
        self._job_result_channels[job_id].put_nowait(result)
        self._job_last_active_time[job_id] = time.monotonic()

    async def _cleanup_task(
        self,
        cleanup_period: int,
        max_idle_time: int,
    ) -> None:
        """Periodically evict job states.

        Args:
            cleanup_period: How often to run the cleanup task, in seconds.
            max_idle_time: Maximum amount of time a job can be idle for, in seconds.
        """
        while True:
            await asyncio.sleep(cleanup_period)
            for job_id in list(self._job_last_active_time.keys()):
                if (
                    time.monotonic() - self._job_last_active_time[job_id]
                    > max_idle_time
                ):
                    self._job_tasks[job_id].cancel()
                    del self._job_infos[job_id]
                    del self._job_rank_infos[job_id]
                    del self._job_result_channels[job_id]
                    del self._job_sched_request_channels[job_id]
                    del self._job_sched_response_channels[job_id]
                    del self._job_tasks[job_id]
                    del self._job_last_active_time[job_id]

    async def _job_task(self, job_id: str, dump_data: bool) -> None:
        """Coalese requests and responses of each rank and interface with the scheduler."""
        result_chan = self._job_result_channels[job_id]
        sched_req_chan = self._job_sched_request_channels[job_id]
        sched_resp_chan = self._job_sched_response_channels[job_id]

        job_info = self._job_infos[job_id]

        try:
            # Wait until all ranks have reported their `RankInfo`s.
            rank_infos = self._job_rank_infos[job_id]
            while True:
                await asyncio.sleep(0.1)
                # Indexing the first element is always safe because this task is
                # created after putting the `RankInfo` of the first-connected rank
                # in `self.job_rank_infos[job_id]`.
                if len(rank_infos) == job_info.world_size:
                    break

            # Sort `RankInfo`s in rank order.
            rank_infos.sort(key=lambda r: r.rank)

            # Create directory to dump PFO server states.
            dump_dir = f"{self.pfo_settings.dump_dir}/{job_id}"
            if dump_data:
                await save_ranks(rank_infos, dump_dir)

            # Instantiate the frequency scheduler.
            scheduler = self.pfo_settings.scheduler(
                job_info,
                rank_infos,
                self.pfo_settings,
                **self.pfo_settings.scheduler_args,
            )

            # Provide next schedules, observe profiling results, and repeat.
            schedule_num = 0
            while True:
                # Compute the next `FrequencySchedule`s.
                schedules = scheduler.next_schedule()

                # Wait until all the ranks ask for the next schedule.
                await asyncio.gather(*[sched_req_chan.get() for _ in rank_infos])

                # Send out `FrequencySchedule`s.
                await asyncio.gather(
                    *[sched_resp_chan[s.rank].put(s) for s in schedules]
                )

                # Gather profiling results from all ranks.
                results = await asyncio.gather(*[result_chan.get() for _ in rank_infos])
                results.sort(key=lambda r: r.rank)

                # Dump profiling results and schedules.
                if dump_data:
                    schedules.sort(key=lambda s: s.rank)
                    await save_prof(results, dump_dir, schedule_num)
                    await save_sched(schedules, dump_dir, schedule_num)

                # Send `ProfilingResult`s to the scheduler.
                scheduler.observe(results)

                # Increment schedule number.
                schedule_num += 1

        except asyncio.CancelledError:
            # This task gets cancelled when it's idle for too long and evicted.
            pass

        except Exception as exc:
            # In case the scheduler errored, send out the exception to the clients.
            # The clients will receive the error when they ask for the next schedule.
            for chan in sched_resp_chan:
                chan.put_nowait(exc)
            raise


def init_global_job_manager(pfo_settings: PFOServerSettings) -> None:
    """Instantiate the global singleton `JobManager`."""
    global GLOBAL_JOB_MANAGER
    GLOBAL_JOB_MANAGER = JobManager(pfo_settings=pfo_settings)


def get_global_job_manager() -> JobManager:
    """Fetch the global singleton `JobManager`."""
    assert GLOBAL_JOB_MANAGER is not None, "`init_global_job_manager` was not called."
    return GLOBAL_JOB_MANAGER



"""Repository for batch size states(Trial, Gaussian Ts arm state)."""

from __future__ import annotations

from collections import defaultdict

from sqlalchemy import and_, select, func
from sqlalchemy.ext.asyncio.session import AsyncSession
from zeus.optimizer.batch_size.server.batch_size_state.commands import (
    CreateTrial,
    ReadTrial,
    UpdateTrial,
)
from zeus.optimizer.batch_size.server.batch_size_state.models import (
    BatchSizeBase,
    ExplorationsPerJob,
    GaussianTsArmState,
    Trial,
    TrialResult,
    TrialResultsPerBs,
)
from zeus.optimizer.batch_size.server.database.repository import DatabaseRepository
from zeus.optimizer.batch_size.server.database.schema import (
    GaussianTsArmStateTable,
    TrialStatus,
    TrialTable,
    TrialType,
)
from zeus.optimizer.batch_size.server.exceptions import ZeusBSOValueError
from zeus.utils.logging import get_logger

logger = get_logger(__name__)


class BatchSizeStateRepository(DatabaseRepository):
    """Repository for handling batch size related operations."""

    def __init__(self, session: AsyncSession):
        """Set db session and intialize fetched trial. We are only updating one trial per session."""
        super().__init__(session)
        self.fetched_trial: TrialTable | None = None
        self.fetched_arm: GaussianTsArmStateTable | None = None

    async def get_next_trial_number(self, job_id: str) -> int:
        """Get next trial number of a given job. Trial number starts from 1 and increase by 1 at a time."""
        stmt = select(func.max(TrialTable.trial_number)).where(
            and_(
                TrialTable.job_id == job_id,
            )
        )
        res = await self.session.scalar(stmt)
        if res is None:
            return 1
        return res + 1

    async def get_trial_results_of_bs(
        self, batch_size: BatchSizeBase, window_size: int
    ) -> TrialResultsPerBs:
        """Load window size amount of results for a given batch size. If window size <= 0, load all of them.

        From all trials, we filter succeeded one since failed/dispatched ones doesn't have a valid result.

        Args:
            batch_size (BatchSizeBase): The batch size object.
            window_size (int): The size of the measurement window.

        Returns:
            TrialResultsPerBs: trial results for the given batch size.
        """
        stmt = (
            select(TrialTable)
            .where(
                and_(
                    TrialTable.job_id == batch_size.job_id,
                    TrialTable.batch_size == batch_size.batch_size,
                    TrialTable.status == TrialStatus.Succeeded,
                )
            )
            .order_by(TrialTable.trial_number.desc())
        )
        if window_size > 0:
            stmt = stmt.limit(window_size)

        res = (await self.session.scalars(stmt)).all()
        return TrialResultsPerBs(
            job_id=batch_size.job_id,
            batch_size=batch_size.batch_size,
            results=[TrialResult.from_orm(t) for t in res],
        )

    async def get_arms(self, job_id: str) -> list[GaussianTsArmState]:
        """Retrieve Gaussian Thompson Sampling arms for a given job.

        Args:
            job_id (str): The ID of the job.

        Returns:
            List[GaussianTsArmStateModel]: List of Gaussian Thompson Sampling arms. These arms are all "good" arms (converged during pruning stage).
            Refer to `GaussianTsArmStateModel`[zeus.optimizer.batch_size.server.batch_size_state.models.GaussianTsArmStateModel] for attributes.
        """
        stmt = select(GaussianTsArmStateTable).where(
            GaussianTsArmStateTable.job_id == job_id
        )
        res = (await self.session.scalars(stmt)).all()
        return [GaussianTsArmState.from_orm(arm) for arm in res]

    async def get_arm(self, bs: BatchSizeBase) -> GaussianTsArmState | None:
        """Retrieve Gaussian Thompson Sampling arm for a given job id and batch size.

        Args:
            bs (BatchSizeBase): The batch size object.

        Returns:
            Optional[GaussianTsArmStateModel]: Gaussian Thompson Sampling arm if found, else None.
            Refer to `GaussianTsArmStateModel`[zeus.optimizer.batch_size.server.batch_size_state.models.GaussianTsArmStateModel] for attributes.
        """
        stmt = select(GaussianTsArmStateTable).where(
            and_(
                GaussianTsArmStateTable.job_id == bs.job_id,
                GaussianTsArmStateTable.batch_size == bs.batch_size,
            )
        )
        arm = await self.session.scalar(stmt)
        if arm is None:
            return None
        self.fetched_arm = arm
        return GaussianTsArmState.from_orm(arm)

    async def get_trial(self, trial: ReadTrial) -> Trial | None:
        """Get a corresponding trial.

        Args:
            trial: job_id, batch_size, trial_number.

        Returns:
            Found Trial. If none found, return None.
        """
        stmt = select(TrialTable).where(
            TrialTable.job_id == trial.job_id,
            TrialTable.batch_size == trial.batch_size,
            TrialTable.trial_number == trial.trial_number,
        )
        fetched_trial = await self.session.scalar(stmt)

        if fetched_trial is None:
            logger.info("get_trial: NoResultFound")
            return None

        self.fetched_trial = fetched_trial
        return Trial.from_orm(fetched_trial)

    def get_trial_from_session(self, trial: ReadTrial) -> Trial | None:
        """Fetch a trial from the session."""
        if (
            self.fetched_trial is None
            or self.fetched_trial.job_id != trial.job_id
            or self.fetched_trial.batch_size != trial.batch_size
            or self.fetched_trial.trial_number != trial.trial_number
        ):
            return None
        return Trial.from_orm(self.fetched_trial)

    def create_trial(self, trial: CreateTrial) -> None:
        """Create a trial in db.

        Refer to `CreateTrial`[zeus.optimizer.batch_size.server.batch_size_state.models.CreateTrial] for attributes.

        Args:
            trial (CreateTrial): The trial to add.
        """
        self.session.add(trial.to_orm())

    def updated_current_trial(self, updated_trial: UpdateTrial) -> None:
        """Update trial in the database (report the result of trial).

        Args:
            updated_trial (UpdateTrial): The updated trial. Refer to `UpdateTrial`[zeus.optimizer.batch_size.server.batch_size_state.models.UpdateTrial] for attributes.
        """
        if self.fetched_trial is None:
            raise ZeusBSOValueError("No trial is fetched.")

        if (
            self.fetched_trial.job_id != updated_trial.job_id
            or self.fetched_trial.batch_size != updated_trial.batch_size
            or self.fetched_trial.trial_number != updated_trial.trial_number
        ):
            raise ZeusBSOValueError("Trying to update invalid trial.")

        self.fetched_trial.end_timestamp = updated_trial.end_timestamp
        self.fetched_trial.status = updated_trial.status
        self.fetched_trial.time = updated_trial.time
        self.fetched_trial.energy = updated_trial.energy
        self.fetched_trial.converged = updated_trial.converged

    def create_arms(self, new_arms: list[GaussianTsArmState]) -> None:
        """Create Gaussian Thompson Sampling arms in the database.

        Args:
            new_arms (List[GaussianTsArmStateModel]): List of new arms to create.
                Refer to `GaussianTsArmStateModel`[zeus.optimizer.batch_size.server.batch_size_state.models.GaussianTsArmStateModel] for attributes.
        """
        self.session.add_all([arm.to_orm() for arm in new_arms])

    def update_arm_state(self, updated_mab_state: GaussianTsArmState) -> None:
        """Update Gaussian Thompson Sampling arm state in db.

        Args:
            updated_mab_state (GaussianTsArmStateModel): The updated arm state.
                Refer to `GaussianTsArmStateModel`[zeus.optimizer.batch_size.server.batch_size_state.models.GaussianTsArmStateModel] for attributes.
        """
        if self.fetched_arm is None:
            raise ZeusBSOValueError("No arm is fetched.")

        if (
            self.fetched_arm.job_id != updated_mab_state.job_id
            or self.fetched_arm.batch_size != updated_mab_state.batch_size
        ):
            raise ZeusBSOValueError(
                "Fetch arm does not correspond with the arm trying to update."
            )

        self.fetched_arm.param_mean = updated_mab_state.param_mean
        self.fetched_arm.param_precision = updated_mab_state.param_precision
        self.fetched_arm.reward_precision = updated_mab_state.reward_precision
        self.fetched_arm.num_observations = updated_mab_state.num_observations

    async def get_explorations_of_job(self, job_id: str) -> ExplorationsPerJob:
        """Retrieve succeeded or ongoing explorations for a given job.

        Args:
            job_id: ID of the job

        Returns:
            ExplorationsPerJob: Explorations for the given batch size.
            Refer to `ExplorationsPerJob`[zeus.optimizer.batch_size.server.batch_size_state.models.ExplorationsPerJob] for attributes.
        """
        stmt = (
            select(TrialTable)
            .where(
                and_(
                    TrialTable.job_id == job_id,
                    TrialTable.type == TrialType.Exploration,
                    TrialTable.status != TrialStatus.Failed,
                )
            )
            .order_by(TrialTable.trial_number.asc())
        )

        explorations = (await self.session.scalars(stmt)).all()
        exps_per_bs: defaultdict[int, list[Trial]] = defaultdict(list)
        for exp in explorations:
            exps_per_bs[exp.batch_size].append(Trial.from_orm(exp))

        return ExplorationsPerJob(job_id=job_id, explorations_per_bs=exps_per_bs)



"""Pipeline frequency optimizer implementation.

The `PipelineFrequencyOptimizer` is to be integrated into the training framework.
It is responsible for communicating with the PFO server and managing
the `FrequencyController` instance, which is responsible for controlling
the frequency of the CPU of the current process.
"""

from __future__ import annotations

import httpx
import torch
import torch.distributed as dist

from zeus.callback import Callback
from zeus.device import get_gpus
from zeus.optimizer.pipeline_frequency.frequency_controller import FrequencyController
from zeus.optimizer.pipeline_frequency.common import (
    GET_FREQUENCY_SCHEDULE_URL,
    REGISTER_JOB_URL,
    REGISTER_RANK_URL,
    JobInfo,
    RankInfo,
    FrequencySchedule,
)
from zeus.utils.framework import sync_execution


class PipelineFrequencyOptimizer(Callback):
    """Pipeline frequency optimizer."""

    def __init__(
        self,
        rank: int,
        dp_rank: int,
        pp_rank: int,
        tp_rank: int,
        device_id: int,
        dp_degree: int,
        pp_degree: int,
        tp_degree: int,
        world_size: int,
        server_url: str,
        job_metadata: str | None = None,
    ) -> None:
        """Initialize the Pipeline frequency optimizer.

        Assumptions:
            - `torch.distributed` has been initialized.
            - `torch.cuda.set_device` has been called with `device_id`.
                This is needed to broadcast the job ID to all ranks.

        The master process (rank 0) will register the job with the Peresus
        server and retrieve the job ID of this job. Then, each rank will
        report itself to the PFO server with the job ID.

        Args:
            rank: Global rank of the current process.
            dp_rank: Rank in the data parallel group.
            pp_rank: Rank in the pipeline parallel group.
            tp_rank: Rank in the tensor parallel group.
            device_id: CUDA device ID that the current process manages.
            dp_degree: Size of the data parallel group.
            pp_degree: Size of the pipeline parallel group.
            tp_degree: Size of the tensor parallel group.
            world_size: Total number of ranks that participate in training.
            server_url: URL of the PFO server.
            job_metadata: An optional arbitrary string that describes the job. This will
                be appended to the job ID if given. Typically for logging purposes.
        """
        if not dist.is_initialized():
            raise RuntimeError(
                "Instantiate `PipelineFrequencyOptimizer` after `init_process_group`."
            )

        self.server_url = server_url
        self.rank = rank
        self.dp_rank = dp_rank
        self.pp_rank = pp_rank
        self.tp_rank = tp_rank
        self.device_id = device_id

        gpus = get_gpus()
        torch.cuda.set_device(device_id)

        # Rank 0 registers the job with the PFO server and retrieves the job ID.
        job_id = None
        if rank == 0:
            job_info = JobInfo(
                pp_degree=pp_degree,
                dp_degree=dp_degree,
                tp_degree=tp_degree,
                world_size=world_size,
                job_metadata=job_metadata,
            )
            response = httpx.post(
                self.server_url + REGISTER_JOB_URL, json=job_info.dict()
            )
            if (code := response.status_code) != 200:
                raise RuntimeError(
                    f"PFO server returned status code {code}: {response.text}"
                )
            job_id = response.json()
            if not isinstance(job_id, str):
                raise RuntimeError(f"PFO server returned a strange job ID: {job_id=}")

        # Rank 0 broadcasts the job ID across all ranks.
        objects = [job_id]
        dist.broadcast_object_list(objects, src=0)
        self.job_id = objects[0]
        if self.job_id is None:
            raise RuntimeError("Failed to broadcast job ID to all ranks")

        # Query the list of available frequencies of the GPU.
        max_mem_freq = max(gpus.getSupportedMemoryClocks(device_id))
        freqs = sorted(
            gpus.getSupportedGraphicsClocks(device_id, max_mem_freq),
            reverse=True,
        )

        # Each rank reports itself to the PFO server with the job ID.
        rank_info = RankInfo(
            rank=self.rank,
            dp_rank=self.dp_rank,
            pp_rank=self.pp_rank,
            tp_rank=self.tp_rank,
            available_frequencies=freqs,
        )
        response = httpx.post(
            self.server_url + REGISTER_RANK_URL.format(job_id=self.job_id),
            json=rank_info.dict(),
        )
        if (code := response.status_code) != 200:
            raise RuntimeError(
                f"PFO server returned status code {code}: {response.text}"
            )

        # The frequency controller is responsible for controlling the frequency
        # of the GPU (device_id) asynchronously.
        self.frequency_controller = FrequencyController(device_id=device_id)

        # Fetch the frequency schedule from the PFO server.
        self.freq_schedule = self._get_frequency_schedule()
        self.freq_schedule_iter = iter(self.freq_schedule)

    def _get_frequency_schedule(self) -> list[tuple[str, int]]:
        """Get the frequency schedule from the PFO server."""
        response = httpx.get(
            self.server_url + GET_FREQUENCY_SCHEDULE_URL.format(job_id=self.job_id),
            params={"rank": self.rank},
            timeout=None,
        )
        if (code := response.status_code) != 200:
            raise RuntimeError(
                f"PFO server returned status code {code}: {response.text}"
            )
        schedule = FrequencySchedule.parse_raw(response.text)
        if schedule.rank != self.rank:
            raise RuntimeError(
                f"PFO server returned a schedule for rank {schedule.rank} to rank {self.rank}"
            )
        return schedule.frequencies

    def on_step_begin(self) -> None:
        """Mark the beginning of a step.

        TODO(jaywonchung): InstructionProfiler iteration start mark.
        """
        pass

    def on_step_end(self) -> None:
        """Mark the end of a step.

        TODO(jaywonchung): InstructionProfiler iteration end mark.
        Also report the profiling result to the PFO server after N iterations.
        """
        # Frequency schedule holds one iteration-worth of frequencies, so at
        # the end of each iteration, the iterator should be exhausted.
        item = next(self.freq_schedule_iter, None)
        if item is not None:
            raise RuntimeError(
                "PFO server returned more frequencies than expected. "
                f"Next expected instruction and frequency is {item}"
            )
        self.freq_schedule_iter = iter(self.freq_schedule)

    def on_instruction_begin(self, name: str) -> None:
        """Mark the beginning of an instruction, like forward and backward.

        Retrieve the next frequency from the schedule, check whether the next
        expected instruction matches the name of the instruction, and set the
        frequency accordingly.
        """
        sync_execution([self.device_id], sync_with="torch")

        # Retrieve the next frequency from the schedule.
        item = next(self.freq_schedule_iter, None)
        if item is None:
            raise RuntimeError("PFO server returned fewer frequencies than expected")

        # Check whether the next expected instruction matches the name of the instruction.
        instruction, frequency = item
        if instruction != name:
            raise RuntimeError(
                f"The next expected instruction is not forward: {instruction}"
            )

        self.frequency_controller.set_frequency(frequency)

    def on_instruction_end(self, name: str) -> None:
        """Mark the end of an instruction, like forward and backward."""



"""Time, energy, and power monitors for Zeus.

The main class of this module is [`ZeusMonitor`][zeus.monitor.energy.ZeusMonitor].

If users wish to monitor power consumption over time, the [`power`][zeus.monitor.power]
module can come in handy.
"""

from zeus.monitor.energy import ZeusMonitor, Measurement
from zeus.monitor.power import PowerMonitor



"""Measure the GPU time and energy consumption of a block of code."""

from __future__ import annotations

import os
import warnings
from typing import Literal
from time import time
from pathlib import Path
from dataclasses import dataclass
from functools import cached_property

from zeus.monitor.power import PowerMonitor
from zeus.utils.logging import get_logger
from zeus.utils.framework import sync_execution as sync_execution_fn
from zeus.device import get_gpus, get_cpus, get_soc
from zeus.device.gpu.common import ZeusGPUInitError, EmptyGPUs
from zeus.device.cpu.common import ZeusCPUInitError, ZeusCPUNoPermissionError, EmptyCPUs
from zeus.device.soc.common import ZeusSoCInitError, SoCMeasurement, EmptySoC

logger = get_logger(__name__)


@dataclass
class Measurement:
    """Measurement result of one window.

    Attributes:
        time: Time elapsed (in seconds) during the measurement window.
        gpu_energy: Maps GPU indices to the energy consumed (in Joules) during the
            measurement window. GPU indices are from the DL framework's perspective
            after applying `CUDA_VISIBLE_DEVICES`.
        cpu_energy: Maps CPU indices to the energy consumed (in Joules) during the measurement
            window. Each CPU index refers to one powerzone exposed by RAPL (intel-rapl:d). This can
            be 'None' if CPU measurement is not available.
        dram_energy: Maps CPU indices to the energy consumed (in Joules) during the measurement
            window. Each CPU index refers to one powerzone exposed by RAPL (intel-rapl:d)  and DRAM
            measurements are taken from sub-packages within each powerzone. This can be 'None' if
            CPU measurement is not available or DRAM measurement is not available.
        soc_energy: If your machine contains an SoC (e.g., Apple silicon), metrics for various
            subsystems of the SoC (e.g., the on-chip CPU, the on-chip GPU) will all be included
            within this field.
    """

    time: float
    gpu_energy: dict[int, float]
    cpu_energy: dict[int, float] | None = None
    dram_energy: dict[int, float] | None = None

    soc_energy: SoCMeasurement | None = None

    @cached_property
    def total_energy(self) -> float:
        """Total energy consumed (in Joules) during the measurement window."""
        # TODO: Update method to total_gpu_energy, which may cause breaking changes in the examples/
        return sum(self.gpu_energy.values())


@dataclass
class MeasurementState:
    """Measurement state to keep track of measurements in start_window.

    Used in ZeusMonitor to map string keys of measurements to this dataclass.

    Attributes:
        time: The beginning timestamp of the measurement window.
        gpu_energy: Maps GPU indices to the energy consumed (in Joules) during the
            measurement window. GPU indices are from the DL framework's perspective
            after applying `CUDA_VISIBLE_DEVICES`.
        cpu_energy: Maps CPU indices to the energy consumed (in Joules) during the measurement
            window. Each CPU index refers to one powerzone exposed by RAPL (intel-rapl:d). This can
            be 'None' if CPU measurement is not available.
        dram_energy: Maps CPU indices to the energy consumed (in Joules) during the measurement
            window. Each CPU index refers to one powerzone exposed by RAPL (intel-rapl:d)  and DRAM
            measurements are taken from sub-packages within each powerzone. This can be 'None' if
            CPU measurement is not available or DRAM measurement is not available.
    """

    time: float
    gpu_energy: dict[int, float]
    cpu_energy: dict[int, float] | None = None
    dram_energy: dict[int, float] | None = None

    @cached_property
    def total_energy(self) -> float:
        """Total energy consumed (in Joules) during the measurement window."""
        return sum(self.gpu_energy.values())


class ZeusMonitor:
    """Measure the GPU energy and time consumption of a block of code.

    Works for multi-GPU and heterogeneous GPU types. Aware of `CUDA_VISIBLE_DEVICES`.
    For instance, if `CUDA_VISIBLE_DEVICES=2,3`, GPU index `1` passed into `gpu_indices`
    will be interpreted as CUDA device `3`.

    You can mark the beginning and end of a measurement window, during which the GPU
    energy and time consumed will be recorded. Multiple concurrent measurement windows
    are supported.

    For Volta or newer GPUs, energy consumption is measured very cheaply with the
    `nvmlDeviceGetTotalEnergyConsumption` API. On older architectures, this API is
    not supported, so a separate Python process is used to poll `nvmlDeviceGetPowerUsage`
    to get power samples over time, which are integrated to compute energy consumption.

    !!! Warning
        Since the monitor may spawn a process to poll the power API on GPUs older than
        Volta, **the monitor should not be instantiated as a global variable
        without guarding it with `if __name__ == "__main__"`**.
        Refer to the "Safe importing of main module" section in the
        [Python documentation](https://docs.python.org/3/library/multiprocessing.html#the-spawn-and-forkserver-start-methods)
        for more details.

    ## Integration Example

    ```python
    import time
    from zeus.monitor import ZeusMonitor

    def training():
        # A dummy training function
        time.sleep(5)

    if __name__ == "__main__":
        # Time/Energy measurements for four GPUs will begin and end at the same time.
        gpu_indices = [0, 1, 2, 3]
        monitor = ZeusMonitor(gpu_indices)

        # Mark the beginning of a measurement window. You can use any string
        # as the window name, but make sure it's unique.
        monitor.begin_window("entire_training")

        # Actual work
        training()

        # Mark the end of a measurement window and retrieve the measurment result.
        result = monitor.end_window("entire_training")

        # Print the measurement result.
        print(f"Training consumed {result.total_energy} Joules.")
        for gpu_idx, gpu_energy in result.gpu_energy.items():
            print(f"GPU {gpu_idx} consumed {gpu_energy} Joules.")
    ```

    Attributes:
        gpu_indices (`list[int]`): Indices of all the CUDA devices to monitor, from the
            DL framework's perspective after applying `CUDA_VISIBLE_DEVICES`.
    """

    def __init__(
        self,
        gpu_indices: list[int] | None = None,
        cpu_indices: list[int] | None = None,
        approx_instant_energy: bool = False,
        log_file: str | Path | None = None,
        sync_execution_with: Literal["torch", "jax"] = "torch",
    ) -> None:
        """Instantiate the monitor.

        Args:
            gpu_indices: Indices of all the CUDA devices to monitor. Time/Energy measurements
                will begin and end at the same time for these GPUs (i.e., synchronized).
                If None, all the GPUs available will be used. `CUDA_VISIBLE_DEVICES`
                is respected if set, e.g., GPU index `1` passed into `gpu_indices` when
                `CUDA_VISIBLE_DEVICES=2,3` will be interpreted as CUDA device `3`.
                `CUDA_VISIBLE_DEVICES`s formatted with comma-separated indices are supported.
            cpu_indices: Indices of the CPU packages to monitor. If None, all CPU packages will
                be used.
            approx_instant_energy: When the execution time of a measurement window is
                shorter than the NVML energy counter's update period, energy consumption may
                be observed as zero. In this case, if `approx_instant_energy` is True, the
                window's energy consumption will be approximated by multiplying the current
                instantaneous power consumption with the window's execution time. This should
                be a better estimate than zero, but it's still an approximation.
            log_file: Path to the log CSV file. If `None`, logging will be disabled.
            sync_execution_with: Deep learning framework to use to synchronize CPU/GPU computations.
                Defaults to `"torch"`, in which case `torch.cuda.synchronize` will be used.
                See [`sync_execution`][zeus.utils.framework.sync_execution] for more details.
        """
        # Save arguments.
        self.approx_instant_energy = approx_instant_energy
        self.sync_with: Literal["torch", "jax"] = sync_execution_with

        # Get GPU instances.
        try:
            self.gpus = get_gpus()
        except ZeusGPUInitError:
            self.gpus = EmptyGPUs()

        # Get CPU instance.
        try:
            self.cpus = get_cpus()
        except ZeusCPUInitError:
            self.cpus = EmptyCPUs()
        except ZeusCPUNoPermissionError as err:
            if cpu_indices:
                raise RuntimeError(
                    "Root privilege is required to read RAPL metrics. See "
                    "https://ml.energy/zeus/getting_started/#system-privileges "
                    "for more information or disable CPU measurement by passing cpu_indices=[] to "
                    "ZeusMonitor"
                ) from err
            self.cpus = EmptyCPUs()

        # Get an SoC instance, if an SoC is present on the host device.
        self.soc_is_present = False
        try:
            self.soc = get_soc()
            self.soc_is_present = True
        except ZeusSoCInitError:
            self.soc = EmptySoC()

        # Resolve GPU indices. If the user did not specify `gpu_indices`, use all available GPUs.
        self.gpu_indices = (
            gpu_indices if gpu_indices is not None else list(range(len(self.gpus)))
        )

        # Resolve CPU indices. If the user did not specify `cpu_indices`, use all available CPUs.
        self.cpu_indices = (
            cpu_indices if cpu_indices is not None else list(range(len(self.cpus)))
        )

        logger.info("Monitoring GPU indices %s.", self.gpu_indices)
        logger.info("Monitoring CPU indices %s", self.cpu_indices)

        # Initialize loggers.
        if log_file is None:
            self.log_file = None
        else:
            if dir := os.path.dirname(log_file):
                os.makedirs(dir, exist_ok=True)
            self.log_file = open(log_file, "w")
            logger.info("Writing measurement logs to %s.", log_file)
            self.log_file.write(
                f"start_time,window_name,elapsed_time,{','.join(map(lambda i: f'gpu{i}_energy', self.gpu_indices))}\n",
            )
            self.log_file.flush()

        # A dictionary that maps the string keys of active measurement windows to
        # the state of the measurement window. Each element in the dictionary is a Measurement State
        # object with:
        #     1) Time elapsed at the beginning of this window.
        #     2) Total energy consumed by each >= Volta GPU at the beginning of
        #        this window (`None` for older GPUs).
        #     3) Total energy consumed by each CPU powerzone at the beginning of this window.
        #        ('None' if CPU measurement is not supported)
        #     4) Total energy consumed by each DRAM in powerzones at the beginning of this window.
        #        ('None' if DRAM measurement is not supported)
        self.measurement_states: dict[str, MeasurementState] = {}

        # Initialize power monitors for older architecture GPUs.
        old_gpu_indices = [
            gpu_index
            for gpu_index in self.gpu_indices
            if not self.gpus.supportsGetTotalEnergyConsumption(gpu_index)
        ]
        if old_gpu_indices:
            self.power_monitor = PowerMonitor(
                gpu_indices=old_gpu_indices, update_period=None
            )
        else:
            self.power_monitor = None

    def _get_instant_power(self) -> tuple[dict[int, float], float]:
        """Measure the power consumption of all GPUs at the current time."""
        power_measurement_start_time: float = time()
        power = {
            i: self.gpus.getInstantPowerUsage(i) / 1000.0 for i in self.gpu_indices
        }
        power_measurement_time = time() - power_measurement_start_time
        return power, power_measurement_time

    def begin_window(self, key: str, sync_execution: bool = True) -> None:
        """Begin a new measurement window.

        Args:
            key: Unique name of the measurement window.
            sync_execution: Whether to wait for asynchronously dispatched computations
                to finish before starting the measurement window. For instance, PyTorch
                and JAX will run GPU computations asynchronously, and waiting them to
                finish is necessary to ensure that the measurement window captures all
                and only the computations dispatched within the window.
        """
        # Make sure the key is unique.
        if key in self.measurement_states:
            raise ValueError(f"Measurement window '{key}' already exists")

        # Synchronize execution (e.g., cudaSynchronize) to freeze at the right time.
        if sync_execution and self.gpu_indices:
            sync_execution_fn(self.gpu_indices, sync_with=self.sync_with)

        # Freeze the start time of the profiling window.
        timestamp: float = time()
        gpu_energy_state: dict[int, float] = {}
        for gpu_index in self.gpu_indices:
            # Query energy directly if the GPU has newer architecture.
            # Otherwise, the Zeus power monitor is running in the background to
            # collect power consumption, so we just need to read the log file later.
            if self.gpus.supportsGetTotalEnergyConsumption(gpu_index):
                gpu_energy_state[gpu_index] = (
                    self.gpus.getTotalEnergyConsumption(gpu_index) / 1000.0
                )

        cpu_energy_state: dict[int, float] = {}
        dram_energy_state: dict[int, float] = {}
        for cpu_index in self.cpu_indices:
            cpu_measurement = self.cpus.getTotalEnergyConsumption(cpu_index) / 1000.0
            cpu_energy_state[cpu_index] = cpu_measurement.cpu_mj
            if cpu_measurement.dram_mj is not None:
                dram_energy_state[cpu_index] = cpu_measurement.dram_mj

        if self.soc_is_present:
            self.soc.beginWindow(key)

        # Add measurement state to dictionary.
        self.measurement_states[key] = MeasurementState(
            time=timestamp,
            gpu_energy=gpu_energy_state,
            cpu_energy=cpu_energy_state or None,
            dram_energy=dram_energy_state or None,
        )
        logger.debug("Measurement window '%s' started.", key)

    def end_window(
        self, key: str, sync_execution: bool = True, cancel: bool = False
    ) -> Measurement:
        """End a measurement window and return the time and energy consumption.

        Args:
            key: Name of an active measurement window.
            sync_execution: Whether to wait for asynchronously dispatched computations
                to finish before starting the measurement window. For instance, PyTorch
                and JAX will run GPU computations asynchronously, and waiting them to
                finish is necessary to ensure that the measurement window captures all
                and only the computations dispatched within the window.
            cancel: Whether to cancel the measurement window. If `True`, the measurement
                window is assumed to be cancelled and discarded. Thus, an empty Measurement
                object will be returned and the measurement window will not be recorded in
                the log file either. `sync_execution` is still respected.
        """
        # Retrieve the start time and energy consumption of this window.
        try:
            measurement_state = self.measurement_states.pop(key)
        except KeyError:
            raise ValueError(f"Measurement window '{key}' does not exist") from None

        # If we're also tracking an SoC, end its window.
        soc_energy_consumption: SoCMeasurement | None = None
        if self.soc_is_present:
            soc_energy_consumption = self.soc.endWindow(key)

        # Take instant power consumption measurements.
        # This, in theory, is introducing extra NVMLs call in the critical path
        # even if computation time is not so short. However, it is reasonable to
        # expect that computation time would be short if the user explicitly
        # turned on the `approx_instant_energy` option. Calling this function
        # as early as possible will lead to more accurate energy approximation.
        power, power_measurement_time = (
            self._get_instant_power() if self.approx_instant_energy else ({}, 0.0)
        )

        # Synchronize execution (e.g., cudaSynchronize) to freeze at the right time.
        if sync_execution and self.gpu_indices:
            sync_execution_fn(self.gpu_indices, sync_with=self.sync_with)

        # If the measurement window is cancelled, return an empty Measurement object.
        if cancel:
            logger.debug("Measurement window '%s' cancelled.", key)

            # If we had a non-None SoC measurement object to report, empty its fields.
            if soc_energy_consumption is not None:
                soc_energy_consumption.zeroAllFields()

            return Measurement(
                time=0.0,
                gpu_energy={gpu: 0.0 for gpu in self.gpu_indices},
                cpu_energy={cpu: 0.0 for cpu in self.cpu_indices},
                soc_energy=soc_energy_consumption,
            )

        end_time: float = time()
        start_time = measurement_state.time
        gpu_start_energy = measurement_state.gpu_energy
        cpu_start_energy = measurement_state.cpu_energy
        dram_start_energy = measurement_state.dram_energy

        time_consumption: float = end_time - start_time
        gpu_energy_consumption: dict[int, float] = {}
        for gpu_index in self.gpu_indices:
            # Query energy directly if the GPU has newer architecture.
            if self.gpus.supportsGetTotalEnergyConsumption(gpu_index):
                end_energy = self.gpus.getTotalEnergyConsumption(gpu_index) / 1000.0
                gpu_energy_consumption[gpu_index] = (
                    end_energy - gpu_start_energy[gpu_index]
                )

        cpu_energy_consumption: dict[int, float] = {}
        dram_energy_consumption: dict[int, float] = {}
        for cpu_index in self.cpu_indices:
            cpu_measurement = self.cpus.getTotalEnergyConsumption(cpu_index) / 1000.0
            if cpu_start_energy is not None:
                cpu_energy_consumption[cpu_index] = (
                    cpu_measurement.cpu_mj - cpu_start_energy[cpu_index]
                )
            if dram_start_energy is not None and cpu_measurement.dram_mj is not None:
                dram_energy_consumption[cpu_index] = (
                    cpu_measurement.dram_mj - dram_start_energy[cpu_index]
                )

        # If there are older GPU architectures, the PowerMonitor will take care of those.
        if self.power_monitor is not None:
            energy = self.power_monitor.get_energy(start_time, end_time)
            # Fallback to the instant power measurement if the PowerMonitor does not
            # have the power samples.
            if energy is None:
                energy = {}
                for gpu_index in self.power_monitor.gpu_indices:
                    energy[gpu_index] = power[gpu_index] * (
                        time_consumption - power_measurement_time
                    )
            gpu_energy_consumption |= energy

        # Approximate energy consumption if the measurement window is too short.
        if self.approx_instant_energy:
            for gpu_index in self.gpu_indices:
                if gpu_energy_consumption[gpu_index] == 0.0:
                    gpu_energy_consumption[gpu_index] = power[gpu_index] * (
                        time_consumption - power_measurement_time
                    )

        # Trigger a warning if energy consumption is zero and approx_instant_energy is not enabled.
        if not self.approx_instant_energy and any(
            energy == 0.0 for energy in gpu_energy_consumption.values()
        ):
            warnings.warn(
                "The energy consumption of one or more GPUs was measured as zero. This means that the time duration of the measurement window was shorter than the GPU's energy counter update period. Consider turning on the `approx_instant_energy` option in `ZeusMonitor`, which approximates the energy consumption of a short time window as instant power draw x window duration.",
                stacklevel=1,
            )

        logger.debug("Measurement window '%s' ended.", key)

        # Add to log file.
        if self.log_file is not None:
            self.log_file.write(
                f"{start_time},{key},{time_consumption},"
                + ",".join(str(gpu_energy_consumption[gpu]) for gpu in self.gpu_indices)
                + "\n"
            )
            self.log_file.flush()

        return Measurement(
            time=time_consumption,
            gpu_energy=gpu_energy_consumption,
            cpu_energy=cpu_energy_consumption or None,
            dram_energy=dram_energy_consumption or None,
            soc_energy=soc_energy_consumption,
        )
